from __future__ import annotations

import math

from iotbx import phil
from scitbx import matrix

from dials.algorithms.shoebox import MaskCode
from dials.array_family import flex

absorption_defs = """
  absorption_correction
    .multiple = True {
    apply = False
      .type = bool
      .help = must be supplied as a user-defined function with a specific interface (not documented)
    algorithm = fuller_kapton kapton_2019 other
      .type = choice
      .help = a specific absorption correction, or implementation thereof \
              kapton_2019 is a more general implementation of fuller_kapton \
              for use on single/multi-panel detectors
    fuller_kapton {
      xtal_height_above_kapton_mm {
        value = 0.02
          .type = float
          .help = height of the beam (or the irradiated crystal) above the kapton tape
        sigma = 0.01
          .type = float
        }
      rotation_angle_deg {
        value = 1.15
          .type = float
          .help = angle of the tape from vertical
        sigma = 0.1
          .type = float
        }
      kapton_half_width_mm {
        value = 1.5875
          .type = float
          .help = forward distance from irradiated crystal to edge of tape nearest detector
        sigma = 0.5
          .type = float
        }
      kapton_thickness_mm {
        value = 0.05
          .type = float
          .help = tape thickness
        sigma = 0.005
          .type = float
        }
      smart_sigmas = False
        .type = bool
        .help = apply spot-specific sigma corrections using kapton param sigmas
      within_spot_sigmas = True
        .type = bool
        .help = calculate initial per-spot sigmas based on variance across pixels in the spot.
        .help = turn this off to get a major speed-up
    }
  }"""

absorption_phil_scope = phil.parse(absorption_defs, process_includes=True)


class get_absorption_correction:
    def __init__(self):
        # Kapton, or polyimide.  C22H10N2O5 Density=1.43, Angle=90.deg
        # Photon Energy (eV), Atten Length (microns)
        data = """6000.00  482.643
   6070.00  500.286
   6140.00  518.362
   6210.00  536.896
   6280.00  555.873
   6350.00  575.302
   6420.00  595.191
   6490.00  615.552
   6560.00  636.382
   6630.00  657.691
   6700.00  679.484
   6770.00  701.758
   6840.00  724.521
   6910.00  747.791
   6980.00  771.561
   7050.00  795.846
   7120.00  820.646
   7190.00  845.963
   7260.00  871.812
   7330.00  898.183
   7400.00  925.082
   7470.00  952.535
   7540.00  980.535
   7610.00  1009.09
   7680.00  1038.18
   7750.00  1067.85
   7820.00  1098.08
   7890.00  1128.88
   7960.00  1160.25
   8030.00  1192.20
   8100.00  1224.76
   8170.00  1257.91
   8240.00  1291.67
   8310.00  1326.01
   8380.00  1360.98
   8450.00  1396.54
   8520.00  1432.72
   8590.00  1469.51
   8660.00  1506.93
   8730.00  1544.96
   8800.00  1583.65
   8870.00  1622.95
   8940.00  1662.90
   9010.00  1703.49
   9080.00  1744.72
   9150.00  1786.59
   9220.00  1829.13
   9290.00  1872.31
   9360.00  1916.16
   9430.00  1960.65
   9500.00  2005.82
   9570.00  2051.65
   9640.00  2098.16
   9710.00  2145.36
   9780.00  2193.22
   9850.00  2241.75
   9920.00  2290.95
   9990.00  2340.86
   10060.0  2391.49
   10130.0  2442.84
   10200.0  2494.86
   10270.0  2547.59
   10340.0  2601.02
   10410.0  2655.14
   10480.0  2709.98
   10550.0  2765.49
   10620.0  2821.73
   10690.0  2878.68
   10760.0  2936.31
   10830.0  2994.67
   10900.0  3053.72
   10970.0  3113.49
   11040.0  3173.96
   11110.0  3235.14
   11180.0  3297.07
   11250.0  3359.67
   11320.0  3423.01
   11390.0  3487.04
   11460.0  3551.76
   11530.0  3617.23
   11600.0  3683.38
   11670.0  3750.23
   11740.0  3817.81
   11810.0  3886.07
   11880.0  3955.05
   11950.0  4024.75
   12020.0  4095.11
   12090.0  4166.20
   12160.0  4237.96
   12230.0  4310.40
   12300.0  4383.60
   12370.0  4457.48
   12440.0  4532.02
   12510.0  4607.25
   12580.0  4683.14
   12650.0  4759.73
   12720.0  4837.01
   12790.0  4914.94
   12860.0  4993.54
   12930.0  5072.79
   13000.0  5152.69"""
        self.energy = flex.double()
        self.microns = flex.double()
        for line in data.split("\n"):
            tokens = line.strip().split()
            self.energy.append(float(tokens[0]))
            self.microns.append(float(tokens[1]))

    def __call__(self, wavelength_ang):
        # calculate energy in eV 12398.425 eV/Ang
        energy_eV = 12398.425 / wavelength_ang
        # interpolate the Henke tables downloaded from lbl.gov
        index_float = (
            (len(self.energy) - 1)
            * (energy_eV - self.energy[0])
            / (self.energy[-1] - self.energy[0])
        )
        fraction, int_idx = math.modf(index_float)
        int_idx = int(int_idx)

        microns = self.microns[int_idx] + fraction * (
            self.microns[int_idx + 1] - self.microns[int_idx]
        )
        return microns / 1000.0


class KaptonAbsorption:
    def __init__(
        self,
        height_mm,
        thickness_mm,
        half_width_mm,
        rotation_angle_deg,
        detector_dist_mm=None,
        pixel_size_mm=None,
        wavelength_ang=None,
        size_fast=None,
        size_slow=None,
        detector=None,
        beam=None,
    ):
        self.height_mm = height_mm  # tool controlled
        self.thickness_mm = thickness_mm  # tool controlled
        self.half_width_mm = half_width_mm  # tool controlled
        self.angle_rad = rotation_angle_deg * math.pi / 180.0  # tool controlled
        self.detector_dist_mm = detector_dist_mm
        self.pixel_size_mm = pixel_size_mm
        self.wavelength_ang = wavelength_ang
        self.size_fast = size_fast
        self.size_slow = size_slow
        self.detector = detector
        self.beam = beam

        self.fast = matrix.col((1, 0, 0))
        self.slow = matrix.col((0, -1, 0))
        if self.beam is None:
            self.beam_direction = matrix.col((0, 0, -1))
        else:
            self.beam_direction = matrix.col(self.beam.get_s0()).normalize()

        # determine absorption coeff (mm-1) through kapton for a given X-ray energy
        G = get_absorption_correction()
        attenuation_length_mm = G(self.wavelength_ang)
        self.abs_coeff = 1 / attenuation_length_mm

        # determine zones of different absorption behavior
        self.surface_normal = matrix.col((-1, 0, 0)).rotate_around_origin(
            axis=self.beam_direction, angle=-self.angle_rad
        )

        # ray tracing
        self.edge_of_tape_normal = self.beam_direction
        self.surface1_point_mm = -self.surface_normal * self.height_mm
        self.surface2_point_mm = -self.surface_normal * (
            self.height_mm + self.thickness_mm
        )
        self.surface3_point_mm = (
            self.surface1_point_mm + self.edge_of_tape_normal * self.half_width_mm
        )
        self.sn1 = self.surface1_point_mm.dot(self.surface_normal)
        self.sn2 = self.surface2_point_mm.dot(self.surface_normal)
        self.sn3 = self.surface3_point_mm.dot(self.edge_of_tape_normal)

    def abs_correction(self, s1):
        try:
            # let's get the unit vector along the direction of the X-ray
            dsurf1 = self.sn1 / (s1.dot(self.surface_normal))
            dsurf2 = self.sn2 / (s1.dot(self.surface_normal))
            dsurf3 = self.sn3 / (s1.dot(self.edge_of_tape_normal))

            # determine path length through kapton tape
            if dsurf3 < dsurf1 or dsurf1 < 0:
                kapton_path_mm = 0.0
            elif dsurf3 < dsurf2:
                kapton_path_mm = dsurf3 - dsurf1
            else:
                kapton_path_mm = dsurf2 - dsurf1

            # determine absorption correction
            absorption_correction = 1 / math.exp(
                -self.abs_coeff * kapton_path_mm
            )  # unitless, >=1
            return absorption_correction
        except ZeroDivisionError:
            return 0

    def abs_correction_flex(self, s1_flex):
        s1_flex_length = len(s1_flex)
        surface_normal = flex.vec3_double(s1_flex_length, self.surface_normal)
        edge_of_tape_normal = flex.vec3_double(s1_flex_length, self.edge_of_tape_normal)

        dsurf1 = flex.double(len(s1_flex), 0)
        dsurf2 = flex.double(len(s1_flex), 0)
        dot_product = s1_flex.dot(surface_normal)
        sel = dot_product != 0
        dsurf1.set_selected(sel, self.sn1 / dot_product.select(sel))
        dsurf2.set_selected(sel, self.sn2 / dot_product.select(sel))
        dsurf3 = self.sn3 / (s1_flex.dot(edge_of_tape_normal))

        # determine path length through kapton tape
        kapton_path_mm = flex.double(s1_flex_length, 0)
        unshadowed_sel = (dsurf3 < dsurf1) | (dsurf1 < 0)
        nearsel = ~unshadowed_sel & (dsurf3 < dsurf2)
        kapton_path_mm.set_selected(nearsel, (dsurf3 - dsurf1).select(nearsel))
        farsel = ~unshadowed_sel & (dsurf3 >= dsurf2)
        kapton_path_mm.set_selected(farsel, (dsurf2 - dsurf1).select(farsel))

        # determine absorption correction
        absorption_correction = 1 / flex.exp(
            -self.abs_coeff * kapton_path_mm
        )  # unitless, >=1
        return absorption_correction

    def abs_bounding_lines(self, as_xy_ints=False):
        if not hasattr(
            self, "segments"
        ):  # avoid recomputing the bounding lines on subsequent calls
            # determine the beam center in mm
            self.beam_ctr_mm_on_detector = matrix.col(
                (
                    self.detector[0].get_beam_centre_px(self.beam_direction)[1]
                    * self.pixel_size_mm,
                    self.detector[0].get_beam_centre_px(self.beam_direction)[0]
                    * self.pixel_size_mm,
                    0,
                )
            )
            # dials format coordinates in pixels of the detector corner
            self.det_zero_mm = (
                self.detector_dist_mm * self.beam_direction
                - self.beam_ctr_mm_on_detector
            )
            # calculate the positions of the boundaries between zones with different absorption behavior
            # get a vector orthogonal to the surface normal and edge normal (along the direction the tape travels)
            self.tape_dir = self.surface_normal.cross(self.edge_of_tape_normal)
            # and one normal to the detector surface
            self.det_normal = self.slow.cross(self.fast)
            # get one more point on the surface of the tape
            self.surface4_point_mm = (
                self.surface3_point_mm - self.surface_normal * self.thickness_mm
            )
            # get the unit vectors normal to the planes of the absorption minimum and maximum
            self.min_normal = self.tape_dir.cross(
                self.surface3_point_mm
            )  # i.e. normal to plane through near edge of tape
            self.max_normal = self.tape_dir.cross(
                self.surface4_point_mm
            )  # i.e. normal to plane through far edge of tape
            self.center_normal_1 = self.tape_dir.cross(
                self.edge_of_tape_normal
            )  # crosshairs part 1
            self.center_normal_2 = self.surface_normal.cross(
                self.edge_of_tape_normal
            )  # crosshairs part 2
            # get some intermediate values (see LaTeX documentation)
            min_det_dot = self.min_normal - self.det_normal
            max_det_dot = self.max_normal - self.det_normal
            # ctr1_det_dot = self.center_normal_1 - self.det_normal # crosshairs part 1
            # ctr2_det_dot = self.center_normal_2 - self.det_normal # crosshairs part 2
            min_max_det_add = self.det_zero_mm.dot(self.det_normal)

            # get the distances along the detector edges to the minimum and maximum absorption edges

            def get_intersection(offset, direction, min_or_max_det_dot, dimension):
                # offset is a number of pixels from the detector origin
                # direction is the fast or slow axis of the detector
                # min_or_max_det_dot is a quantity to be used in a dot product (see documentation)
                # dimension is the number of pixels in the panel along the direction specified
                try:
                    num = (
                        -(self.det_zero_mm + offset * self.pixel_size_mm).dot(
                            min_or_max_det_dot
                        )
                        - min_max_det_add
                    )
                    denom = direction.dot(min_or_max_det_dot)
                    dist = num / denom
                    if (
                        0 <= dist and dist <= dimension * self.pixel_size_mm
                    ):  # point is on a detector edge
                        return (
                            self.det_zero_mm
                            + offset * self.pixel_size_mm
                            + dist * direction
                        )
                    else:
                        return None
                except ZeroDivisionError:
                    return None

            zero = matrix.col((0, 0, 0))
            self.segments = []
            for edge in [min_det_dot, max_det_dot]:
                # for edge in [min_det_dot, max_det_dot, ctr1_det_dot, ctr2_det_dot]: # crosshairs
                # calculate all the possible intersections on the detector edges
                d_top_fast = get_intersection(zero, self.fast, edge, self.size_fast)
                d_bottom_fast = get_intersection(
                    self.size_slow * self.slow, self.fast, edge, self.size_fast
                )
                d_left_slow = get_intersection(zero, self.slow, edge, self.size_slow)
                d_right_slow = get_intersection(
                    self.size_fast * self.fast, self.slow, edge, self.size_slow
                )
                # collect the ones that aren't None
                points = set()
                for p in [d_top_fast, d_bottom_fast, d_left_slow, d_right_slow]:
                    if p is not None:
                        points.add(p)
                assert len(points) <= 2, (
                    "There seems to be a plane intersection at a corner of the detector "
                    + "that is registering as an intersection on both the fast and the slow axes with slightly different "
                    + "coordinates. Please consider figuring out how to discard one of the two."
                )
                assert len(points) == 2, "The edges of the detector were not found."
                self.segments.append(tuple(points))
        if as_xy_ints:
            if not hasattr(self, "segments_as_xy"):
                self.segments_as_xy = []
                for seg in self.segments:
                    seg_as_xy = []
                    for point in seg:
                        as_xy = point - self.det_zero_mm
                        x_int = int(as_xy.dot(self.slow) / self.pixel_size_mm)
                        y_int = int(as_xy.dot(self.fast) / self.pixel_size_mm)
                        seg_as_xy += [x_int, y_int]
                    self.segments_as_xy.append(tuple(seg_as_xy))
            return self.segments_as_xy
        else:
            return self.segments


class image_kapton_correction:
    def __init__(
        self,
        panel_size_px=None,
        pixel_size_mm=None,
        detector_dist_mm=None,
        wavelength_ang=None,
        reflections_sele=None,
        params=None,
        expt=None,
        refl=None,
        smart_sigmas=True,
        logger=None,
    ):
        self.panel_size_px = panel_size_px
        self.pixel_size_mm = pixel_size_mm
        self.detector_dist_mm = detector_dist_mm
        self.wavelength_ang = wavelength_ang
        self.reflections_sele = reflections_sele
        self.params = params
        self.expt = expt
        self.refl = refl
        self.smart_sigmas = smart_sigmas
        self.logger = logger
        self.extract_params()

    def extract_params(self):
        h = self.params.xtal_height_above_kapton_mm.value
        t = self.params.kapton_thickness_mm.value
        w = self.params.kapton_half_width_mm.value
        a = self.params.rotation_angle_deg.value
        self.kapton_params = (h, t, w, a)

        if self.smart_sigmas:
            sig_h = self.params.xtal_height_above_kapton_mm.sigma
            sig_t = self.params.kapton_thickness_mm.sigma
            sig_w = self.params.kapton_half_width_mm.sigma
            sig_a = self.params.rotation_angle_deg.sigma
            self.kapton_params_sigmas = (sig_h, sig_t, sig_w, sig_a)
            assert all(
                sig >= 0 for sig in self.kapton_params_sigmas
            ), "Kapton param sigmas must be nonnegative"
            self.kapton_params_maxes = [
                [
                    self.kapton_params[i] + self.kapton_params_sigmas[j]
                    if j == i
                    else self.kapton_params[i]
                    for i in range(4)
                ]
                for j in range(4)
            ]
            self.kapton_params_mins = [
                [
                    max(self.kapton_params[i] - self.kapton_params_sigmas[j], 0.001)
                    if j == i
                    else self.kapton_params[i]
                    for i in range(3)
                ]
                + [a]
                for j in range(3)
            ] + [[self.kapton_params[i] for i in range(3)] + [a - sig_a]]

    def __call__(self, plot=False):
        def correction_and_within_spot_sigma(params_version, variance_within_spot=True):
            absorption = KaptonAbsorption(
                params_version[0],
                params_version[1],
                params_version[2],
                params_version[3],
                self.detector_dist_mm,
                self.pixel_size_mm,
                self.wavelength_ang,
                *map(float, self.panel_size_px),
            )
            detector = self.expt.detector
            # y_max = int(detector[0].millimeter_to_pixel(detector[0].get_image_size())[1])

            absorption_corrections = flex.double()
            absorption_sigmas = (
                flex.double()
            )  # std dev of corrections for pixels within a spot, default sigma

            if variance_within_spot:
                mask_code = MaskCode.Foreground | MaskCode.Valid
                for iref in range(len(self.reflections_sele)):
                    kapton_correction_vector = flex.double()
                    # foreground: integration mask
                    shoebox = self.reflections_sele[iref]["shoebox"]
                    foreground = (
                        (shoebox.mask.as_1d() & mask_code) == mask_code
                    ).iselection()
                    width = shoebox.xsize()
                    fast_coords = (foreground % width).as_int()  # within spot
                    slow_coords = (foreground / width).as_int()  # within spot
                    f_absolute = fast_coords + shoebox.bbox[0]  # relative to detector
                    s_absolute = slow_coords + shoebox.bbox[2]  # relative to detector
                    lab_coords = detector[0].get_lab_coord(
                        detector[0].pixel_to_millimeter(
                            flex.vec2_double(
                                f_absolute.as_double(), s_absolute.as_double()
                            )
                        )
                    )
                    s1 = lab_coords.each_normalize()
                    kapton_correction_vector.extend(absorption.abs_correction_flex(s1))
                    average_kapton_correction = flex.mean(kapton_correction_vector)
                    absorption_corrections.append(average_kapton_correction)
                    try:
                        spot_px_stddev = flex.mean_and_variance(
                            kapton_correction_vector
                        ).unweighted_sample_standard_deviation()
                    except Exception:
                        assert (
                            len(kapton_correction_vector) == 1
                        ), "stddev could not be calculated"
                        spot_px_stddev = 0
                    absorption_sigmas.append(spot_px_stddev)
                return absorption_corrections, absorption_sigmas
            else:
                s1_flex = self.reflections_sele["s1"].each_normalize()
                absorption_corrections = absorption.abs_correction_flex(s1_flex)
                return absorption_corrections, None

        # loop through modified Kapton parameters to get alternative corrections and estimate sigmas as
        # maximum variation between these versions of the corrections, on a per-spot basis, or the standard
        # deviation within a single spot, whichever is larger.
        self.logger.info("Calculating kapton corrections to integrated intensities...")
        corrections, sigmas = correction_and_within_spot_sigma(
            self.kapton_params, variance_within_spot=self.params.within_spot_sigmas
        )
        if self.smart_sigmas:
            for p in self.kapton_params_mins + self.kapton_params_maxes:
                self.logger.info("Calculating smart sigmas...")
                modif_corrections, _ = correction_and_within_spot_sigma(
                    p, variance_within_spot=False
                )
                perturbed = flex.abs(corrections - modif_corrections)
                if sigmas is None:
                    sigmas = perturbed
                else:
                    replace_sel = perturbed > sigmas
                    sigmas.set_selected(replace_sel, perturbed.select(replace_sel))
        if plot:
            from matplotlib import pyplot as plt

            for (title, data) in [("corrections", corrections), ("sigmas", sigmas)]:
                plt.hist(data, 20)
                plt.title(title)
                plt.show()
        if self.logger is not None:
            self.logger.info(
                "Returning absorption corrections and sigmas for %d spots"
                % len(corrections)
            )
        return corrections, sigmas


class multi_kapton_correction:
    def __init__(self, experiments, integrated, kapton_params, logger=None):
        self.experiments = experiments
        self.reflections = integrated
        self.params = kapton_params
        self.logger = logger

    def __call__(self):
        self.corrected_reflections = flex.reflection_table()
        for expt, refl in zip(
            self.experiments, self.reflections.split_by_experiment_id()
        ):
            # extract experiment details
            detector = expt.detector
            assert len(detector) == 1  # for now
            panel = detector[0]
            panel_size_px = panel.get_image_size()
            pixel_size_mm = panel.get_pixel_size()[0]
            detector_dist_mm = panel.get_distance()
            beam = expt.beam
            wavelength_ang = beam.get_wavelength()

            # exclude reflections with no foreground pixels
            refl_valid = refl.select(
                refl["num_pixels.valid"] > 0 and refl["num_pixels.foreground"] > 0
            )
            refl_zero = refl_valid.select(refl_valid["intensity.sum.value"] == 0)
            refl_nonzero = refl_valid.select(refl_valid["intensity.sum.value"] != 0)

            def correct(refl_sele, smart_sigmas=True):
                kapton_correction = image_kapton_correction(
                    panel_size_px=panel_size_px,
                    pixel_size_mm=pixel_size_mm,
                    detector_dist_mm=detector_dist_mm,
                    wavelength_ang=wavelength_ang,
                    reflections_sele=refl_sele,
                    params=self.params,
                    expt=expt,
                    refl=refl,
                    smart_sigmas=smart_sigmas,
                    logger=self.logger,
                )

                k_corr, k_sigmas = kapton_correction()
                refl_sele["kapton_absorption_correction"] = k_corr
                if smart_sigmas:
                    refl_sele["kapton_absorption_correction_sigmas"] = k_sigmas
                    # apply corrections and propagate error
                    # term1 = (sig(C)/C)^2
                    # term2 = (sig(Imeas)/Imeas)^2
                    # I' = C*I
                    # sig^2(I') = (I')^2*(term1 + term2)
                    integrated_data = refl_sele["intensity.sum.value"]
                    integrated_variance = refl_sele["intensity.sum.variance"]
                    integrated_sigma = flex.sqrt(integrated_variance)
                    term1 = flex.pow(k_sigmas / k_corr, 2)
                    term2 = flex.pow(integrated_sigma / integrated_data, 2)
                    integrated_data *= k_corr
                    integrated_variance = flex.pow(integrated_data, 2) * (term1 + term2)
                    refl_sele["intensity.sum.value"] = integrated_data
                    refl_sele["intensity.sum.variance"] = integrated_variance
                    # order is purposeful: the two lines above require that integrated_data
                    # has already been corrected!
                else:
                    refl_sele["intensity.sum.value"] *= k_corr
                    refl_sele["intensity.sum.variance"] *= flex.pow2(k_corr)
                return refl_sele

            if len(refl_zero) > 0 and self.params.smart_sigmas:
                # process nonzero intensity reflections with smart sigmas as requested
                # but turn them off for zero intensity reflections to avoid a division by zero
                # during error propagation. Not at all certain this is the best way.
                self.corrected_reflections.extend(
                    correct(refl_nonzero, smart_sigmas=True)
                )
                self.corrected_reflections.extend(
                    correct(refl_zero, smart_sigmas=False)
                )
            else:
                self.corrected_reflections.extend(
                    correct(refl_valid, smart_sigmas=self.params.smart_sigmas)
                )

        return self.experiments, self.corrected_reflections



from __future__ import annotations


def test():
    from dials.algorithms.profile_model.modeller import (
        MultiExpProfileModeller,
        ProfileModellerIface,
    )
    from dials.array_family import flex

    class Modeller(ProfileModellerIface):
        def __init__(self, index, expected):
            self.index = index
            self.accumulated = False
            self.finalized = False
            self.expected = expected
            super().__init__()

        def model(self, reflections):
            assert reflections["id"].all_eq(self.index)
            assert len(reflections) == self.expected

        def accumulate(self, other):
            self.accumulated = True
            assert self.index == other.index

        def finalize(self):
            assert self.accumulated
            self.finalized = True

    # The expected number of reflections
    expected = [100, 200, 300, 400, 500]

    # Create some reflections
    reflections = flex.reflection_table()
    reflections["id"] = flex.int()
    for idx in range(len(expected)):
        for n in range(expected[idx]):
            reflections.append({"id": idx})

    # Create two modellers
    modeller1 = MultiExpProfileModeller()
    modeller2 = MultiExpProfileModeller()
    for idx in range(len(expected)):
        modeller1.add(Modeller(idx, expected[idx]))
        modeller2.add(Modeller(idx, expected[idx]))

    # Model the reflections
    modeller1.model(reflections)
    modeller2.model(reflections)

    # Accumulate
    modeller1.accumulate(modeller2)

    # Finalize
    modeller1.finalize()

    # Check finalized
    assert modeller1.finalized



from __future__ import annotations

from dials.array_family import flex


class ValidatedMultiExpProfileModeller:
    """
    A class to wrap profile modeller for validation
    """

    def __init__(self, modellers):
        """
        Init the list of MultiExpProfileModeller modellers
        """
        self.modellers = modellers
        self.finalized_modeller = None

    def __getitem__(self, index):
        """
        Get a modeller
        """
        return self.modellers[index]

    def model(self, reflections):
        """
        Do the modelling for all modellers
        """

        if "profile.index" not in reflections:
            assert len(self.modellers) == 1
            self.modellers[0].model(reflections)
        else:
            for i, modeller in enumerate(self.modellers):
                mask = reflections["profile.index"] != i
                indices = flex.size_t(range(len(mask))).select(mask)
                if len(indices) > 0:
                    subsample = reflections.select(indices)
                    modeller.model(subsample)
                    reflections.set_selected(indices, subsample)

    def validate(self, reflections):
        """
        Do the validation.
        """

        results = []
        for i, modeller in enumerate(self.modellers):
            mask = reflections["profile.index"] != i
            indices = flex.size_t(range(len(mask))).select(mask)
            if len(indices) > 0:
                subsample = reflections.select(indices)
                modeller.validate(subsample)
                reflections.set_selected(indices, subsample)
                corr = subsample["profile.correlation"]
                mean_corr = flex.mean(corr)
            else:
                mean_corr = None
            results.append(mean_corr)
        return results

    def accumulate(self, other):
        """
        Accumulate the modellers
        """
        assert len(self) == len(other)
        for ms, mo in zip(self, other):
            ms.accumulate(mo)

    def finalize(self):
        """
        Finalize the model
        """
        assert not self.finalized()
        for m in self:
            if self.finalized_modeller is None:
                self.finalized_modeller = m.copy()
            else:
                self.finalized_modeller.accumulate(m)
            m.finalize()
        self.finalized_modeller.finalize()

    def finalized(self):
        """
        Check if the model has been finalized.
        """
        return self.finalized_modeller is not None

    def finalized_model(self):
        """
        Get the finalized model
        """
        assert self.finalized
        return self.finalized_modeller

    def __iter__(self):
        """
        Iterate through the modellers
        """
        yield from self.modellers

    def __len__(self):
        """
        Return the number of modellers
        """
        return len(self.modellers)



from __future__ import annotations

import sys
import time

from dials.util import debug_console

interactive_console = debug_console


class ProgressBarTimer:
    """A simple timer for the progress bar."""

    def __init__(self):
        """Init the progress bar timer."""
        self._last_time = time.time()
        self._last_perc = 0
        self._update_period = 0.5
        self._n_seconds_left = -1

    def get_elapsed_time(self):
        return time.time() - self._last_time

    def update(self, percent):
        """Update the timer."""
        # Get the current time diff between last time
        curr_time = time.time()
        diff_time = curr_time - self._last_time

        # Only update after certain period or at 100%
        if percent < 0:
            percent = 0
        if percent > 100:
            percent = 100
        if diff_time >= self._update_period or percent >= 100:

            # Check the difference in percentage and calculate
            # number of seconds remaining
            diff_perc = percent - self._last_perc
            if diff_perc == 0:
                self._n_seconds_left = 0
            else:
                self._n_seconds_left = diff_time * (100 - percent) / diff_perc

        # Return number of seconds
        return self._n_seconds_left


class ProgressBar:
    """A command line progress bar."""

    def __init__(
        self,
        title=None,
        spinner=True,
        bar=True,
        estimate_time=True,
        indent=0,
        length=80,
    ):
        """Init the progress bar parameters."""

        # Set the parameters
        self._title = title
        self._indent = indent
        self._spinner = spinner
        self._estimate_time = estimate_time
        self._bar = bar
        self._length = length

        self._timer = ProgressBarTimer()

        # Print 0 percent
        self.update(0)

    def update(self, fpercent):
        """Update the progress bar with a percentage."""
        from math import ceil

        # do not update if not a tty
        if not sys.stdout.isatty():
            return

        # Get integer percentage
        percent = int(fpercent)
        if percent < 0:
            percent = 0
        if percent > 100:
            percent = 100

        # Add a percentage counter
        right_str = ""
        left_str = ""
        if sys.stdout.isatty():
            left_str = "\r"
        left_str += " " * self._indent

        # Add a title if given
        if self._title:
            left_str += self._title + ": "

        left_str += f"{percent: >3}%"

        # Add a spinner
        if self._spinner:
            left_str += " "
            left_str += "[ {} ]".format(r"-\|/"[percent % 4])

        # Add a timer
        if self._estimate_time:
            n_seconds_left = self._timer.update(fpercent)
            if n_seconds_left < 0:
                n_seconds_left = "?"
            else:
                n_seconds_left = int(ceil(n_seconds_left))
            right_str = " " + f"est: {n_seconds_left}s" + right_str

        # Add a bar
        if self._bar:
            bar_length = self._length - (len(left_str) + len(right_str)) - 5
            n_char = int(percent * bar_length / 100)
            n_space = bar_length - n_char
            left_str += " "
            left_str += f"[ {'=' * n_char}>{' ' * n_space} ]"

        # Append strings
        progress_str = left_str + right_str

        # Print progress string to stdout
        sys.stdout.write(progress_str)
        sys.stdout.flush()

    def finished(self, string=None):
        """The progress bar is finished."""
        if string:
            self._title = string
        else:
            string = ""
        """ Print the 'end of comand' string."""
        if self._estimate_time:
            # Get the time string
            time_string = f"{self._timer.get_elapsed_time():.2f}s"

            # Truncate the string
            max_length = self._length - self._indent - len(time_string) - 1
            string = string[:max_length]

            # Add an indent and a load of dots and then the time string
            dot_length = 1 + max_length - len(string)
            string = (" " * self._indent) + string
            string = string + "." * (dot_length)
            string = string + time_string

        else:

            # Truncate the string
            max_length = self._length - self._indent
            string = string[:max_length]

            # Add a load of dots
            dot_length = max_length - len(string)
            string = (" " * self._indent) + string
            string = string + "." * (dot_length)

        # Write the string to stdout
        if sys.stdout.isatty():
            string = "\r" + string + "\n"
        else:
            string = string + "\n"
        sys.stdout.write(string)
        sys.stdout.flush()


class Command:
    """Class to nicely print out a command with timing info."""

    # Variables available in class methods
    indent = 0
    max_length = 80
    print_time = True

    @classmethod
    def start(cls, string):
        """Print the 'start command' string."""
        # from termcolor import colored

        # Get the command start time
        cls._start_time = time.time()

        # do not output if not a tty
        if not sys.stdout.isatty():
            return

        # Truncate the string to the maximum length
        max_length = cls.max_length - cls.indent - 3
        string = string[:max_length]
        string = (" " * cls.indent) + string + "..."

        # Write the string to stdout
        sys.stdout.write(string)
        sys.stdout.flush()

    @classmethod
    def end(cls, string):
        """Print the 'end of command' string."""
        # from termcolor import colored

        # Check if we want to print the time or not
        if cls.print_time:

            # Get the time string
            time_string = f"{time.time() - cls._start_time:.2f}s"

            # Truncate the string
            max_length = cls.max_length - cls.indent - len(time_string) - 1
            string = string[:max_length]

            # Add an indent and a load of dots and then the time string
            dot_length = 1 + max_length - len(string)
            string = (" " * cls.indent) + string
            string = string + "." * (dot_length)
            string = string + time_string

        else:

            # Truncate the string
            max_length = cls.max_length - cls.indent
            string = string[:max_length]

            # Add a load of dots
            dot_length = max_length - len(string)
            string = (" " * cls.indent) + string
            string = string + "." * (dot_length)

        # Write the string to stdout
        if sys.stdout.isatty():
            string = "\r" + string + "\n"
        else:
            string = string + "\n"
        sys.stdout.write(string)
        sys.stdout.flush()


try:
    import termcolor
except ImportError:
    termcolor = None


def coloured(text, *args, **kwargs):
    if not sys.stdout.isatty() or termcolor is None:
        return text
    return termcolor.colored(text, *args, **kwargs)


def heading(text):
    return coloured(text, attrs=["bold"])


if __name__ == "__main__":
    p = ProgressBar()

    for j in range(100):
        p.update(j)
        time.sleep(0.05)

    p.finished()

    Command.start("Starting to do a command")
    time.sleep(1)
    Command.end("Ending the command")



from __future__ import annotations

import random

import numpy as np

from cctbx import sgtbx

from dials.algorithms.clustering.unit_cell import cluster_unit_cells


def test_unit_cell():
    # generate some random unit cells
    sgi = sgtbx.space_group_info("P1")
    crystal_symmetries = [
        sgi.any_compatible_crystal_symmetry(volume=random.uniform(990, 1010))
        for i in range(10)
    ]
    lattice_ids = list(range(len(crystal_symmetries)))
    result = cluster_unit_cells(crystal_symmetries, lattice_ids)
    assert "cluster_1" in str(result)
    assert len(result.clusters) == 1
    assert "dcoord" in result.dendrogram.keys()
    assert isinstance(result.linkage_matrix, np.ndarray)



from __future__ import annotations

import pytest

import scitbx.matrix
from cctbx import crystal, sgtbx, uctbx
from cctbx.sgtbx import bravais_types
from dxtbx.model import Crystal

from dials.algorithms.indexing import symmetry


@pytest.mark.parametrize("space_group_symbol", bravais_types.acentric)
def test_SymmetryHandler(space_group_symbol):

    sgi = sgtbx.space_group_info(symbol=space_group_symbol)
    sg = sgi.group()
    cs = sgi.any_compatible_crystal_symmetry(volume=10000)
    uc = cs.unit_cell()

    handler = symmetry.SymmetryHandler(unit_cell=uc, space_group=sg)

    assert (
        handler.target_symmetry_primitive.space_group()
        == sg.build_derived_patterson_group().info().primitive_setting().group()
    )
    assert (
        handler.target_symmetry_reference_setting.space_group()
        == sg.build_derived_patterson_group().info().reference_setting().group()
    )

    # test apply_symmetry on the primitive setting
    cs_primitive = cs.primitive_setting()
    B = scitbx.matrix.sqr(
        cs_primitive.unit_cell().fractionalization_matrix()
    ).transpose()
    crystal = Crystal(B, sgtbx.space_group())
    crystal_new, cb_op = handler.apply_symmetry(crystal)
    crystal_new.get_crystal_symmetry(assert_is_compatible_unit_cell=True)

    # test apply_symmetry on the minimum cell setting
    cs_min_cell = cs.minimum_cell()
    B = scitbx.matrix.sqr(
        cs_min_cell.unit_cell().fractionalization_matrix()
    ).transpose()
    crystal = Crystal(B, sgtbx.space_group())
    crystal_new, cb_op = handler.apply_symmetry(crystal)
    crystal_new.get_crystal_symmetry(assert_is_compatible_unit_cell=True)

    handler = symmetry.SymmetryHandler(space_group=sg)
    assert handler.target_symmetry_primitive.unit_cell() is None
    assert (
        handler.target_symmetry_primitive.space_group()
        == sg.build_derived_patterson_group().info().primitive_setting().group()
    )
    assert handler.target_symmetry_reference_setting.unit_cell() is None
    assert (
        handler.target_symmetry_reference_setting.space_group()
        == sg.build_derived_patterson_group().info().reference_setting().group()
    )

    # test apply_symmetry on the primitive setting
    cs_primitive = cs.primitive_setting()
    B = scitbx.matrix.sqr(
        cs_primitive.unit_cell().fractionalization_matrix()
    ).transpose()
    crystal = Crystal(B, sgtbx.space_group())
    crystal_new, cb_op = handler.apply_symmetry(crystal)
    crystal_new.get_crystal_symmetry(assert_is_compatible_unit_cell=True)

    handler = symmetry.SymmetryHandler(
        unit_cell=cs_min_cell.unit_cell(),
        space_group=sgtbx.space_group(),
    )
    assert handler.target_symmetry_primitive.unit_cell().volume() == pytest.approx(
        cs_min_cell.unit_cell().volume()
    )
    assert handler.target_symmetry_primitive.space_group() == sgtbx.space_group("P-1")
    assert (
        handler.target_symmetry_reference_setting.unit_cell().volume()
        == pytest.approx(cs_min_cell.unit_cell().volume())
    )
    assert handler.target_symmetry_reference_setting.space_group() == sgtbx.space_group(
        "P-1"
    )


# https://github.com/dials/dials/issues/1254
def test_SymmetryHandler_no_match():
    sgi = sgtbx.space_group_info(symbol="P422")
    cs = sgi.any_compatible_crystal_symmetry(volume=10000)
    B = scitbx.matrix.sqr(cs.unit_cell().fractionalization_matrix()).transpose()
    crystal = Crystal(B, sgtbx.space_group())

    handler = symmetry.SymmetryHandler(
        unit_cell=None, space_group=sgtbx.space_group_info("I23").group()
    )
    assert handler.apply_symmetry(crystal) == (None, None)


# https://github.com/dials/dials/issues/1217
@pytest.mark.parametrize(
    "crystal_symmetry",
    [
        crystal.symmetry(
            unit_cell=(
                44.66208171,
                53.12629403,
                62.53397661,
                64.86329707,
                78.27343894,
                90,
            ),
            space_group_symbol="C 1 2/m 1 (z,x+y,-2*x)",
        ),
        crystal.symmetry(
            unit_cell=(44.3761, 52.5042, 61.88555952, 115.1002877, 101.697107, 90),
            space_group_symbol="C 1 2/m 1 (-z,x+y,2*x)",
        ),
    ],
)
def test_symmetry_handler_c2_i2(crystal_symmetry):
    cs_ref = crystal_symmetry.as_reference_setting()
    cs_ref = cs_ref.change_basis(
        cs_ref.change_of_basis_op_to_best_cell(best_monoclinic_beta=False)
    )
    cs_best = cs_ref.best_cell()
    # best -> ref is different to cs_ref above
    cs_best_ref = cs_best.as_reference_setting()
    assert not cs_ref.is_similar_symmetry(cs_best_ref)

    B = scitbx.matrix.sqr(
        crystal_symmetry.unit_cell().fractionalization_matrix()
    ).transpose()
    cryst = Crystal(B, sgtbx.space_group())

    for cs in (crystal_symmetry, cs_ref, cs_best):
        print(cs)
        handler = symmetry.SymmetryHandler(space_group=cs.space_group())
        new_cryst, cb_op = handler.apply_symmetry(cryst)
        assert (
            new_cryst.change_basis(cb_op).get_crystal_symmetry().is_similar_symmetry(cs)
        )

    for cs in (crystal_symmetry, cs_ref, cs_best, cs_best_ref):
        print(cs)
        handler = symmetry.SymmetryHandler(
            unit_cell=cs.unit_cell(), space_group=cs.space_group()
        )
        new_cryst, cb_op = handler.apply_symmetry(cryst)
        assert (
            new_cryst.change_basis(cb_op).get_crystal_symmetry().is_similar_symmetry(cs)
        )


crystal_symmetries = []
cs = crystal.symmetry(
    unit_cell=uctbx.unit_cell("76, 115, 134, 90, 99.07, 90"),
    space_group_info=sgtbx.space_group_info(symbol="I2"),
)
crystal_symmetries.append(
    crystal.symmetry(
        unit_cell=cs.minimum_cell().unit_cell(), space_group=sgtbx.space_group()
    )
)
cs = crystal.symmetry(
    unit_cell=uctbx.unit_cell("42,42,40,90,90,90"),
    space_group_info=sgtbx.space_group_info(symbol="P41212"),
)
crystal_symmetries.append(cs.change_basis(sgtbx.change_of_basis_op("c,a,b")))

for symbol in bravais_types.acentric:
    sgi = sgtbx.space_group_info(symbol=symbol)
    cs = crystal.symmetry(
        unit_cell=sgi.any_compatible_unit_cell(volume=1000), space_group_info=sgi
    )
    cs = cs.niggli_cell().as_reference_setting().primitive_setting()
    crystal_symmetries.append(cs)


@pytest.mark.parametrize("crystal_symmetry", crystal_symmetries)
def test_find_matching_symmetry(crystal_symmetry):

    cs = crystal_symmetry
    cs.show_summary()

    for op in ("x,y,z", "z,x,y", "y,z,x", "-x,z,y", "y,x,-z", "z,-y,x")[:]:
        cb_op = sgtbx.change_of_basis_op(op)

        uc_inp = cs.unit_cell().change_basis(cb_op)

        for ref_uc, ref_sg in [
            (cs.unit_cell(), cs.space_group()),
            (None, cs.space_group()),
        ][:]:

            best_subgroup = symmetry.find_matching_symmetry(
                uc_inp, target_space_group=ref_sg
            )
            cb_op_inp_best = best_subgroup["cb_op_inp_best"]

            assert uc_inp.change_basis(cb_op_inp_best).is_similar_to(
                cs.as_reference_setting().best_cell().unit_cell()
            )



from __future__ import annotations

from dials.algorithms.profile_model.gaussian_rs.calculator import (
    _select_reflections_for_sigma_calc,
)
from dials.array_family import flex


def test_select_reflections_for_sigma_calc():
    """Test the reflection selection helper function."""
    # first select all if below threshold
    reflections = flex.reflection_table()
    reflections["id"] = flex.int(range(0, 1000))
    reflections.set_flags(flex.bool(1000, True), reflections.flags.used_in_refinement)
    reflections = _select_reflections_for_sigma_calc(reflections, 10000)
    assert reflections.size() == 1000

    # select used_in_refinement if not all used in refinement and above threshold
    reflections = flex.reflection_table()
    reflections["id"] = flex.int(range(0, 1000))
    good = flex.bool(1000, False)
    sel = flex.size_t(i for i in range(0, 1000, 2))
    good.set_selected(sel, True)
    reflections.set_flags(good, reflections.flags.used_in_refinement)
    reflections = _select_reflections_for_sigma_calc(
        reflections, min_number_of_refl=200
    )
    assert reflections.size() == 500
    assert list(reflections["id"])[0:50] == list(range(0, 100, 2))

    # top up if not enough used in refinement compared to threshold
    reflections = flex.reflection_table()
    reflections["id"] = flex.int(range(0, 1000))
    good = flex.bool(1000, False)
    sel = flex.size_t(i for i in range(0, 1000, 2))
    good.set_selected(sel, True)
    reflections.set_flags(good, reflections.flags.used_in_refinement)
    reflections = _select_reflections_for_sigma_calc(
        reflections, min_number_of_refl=700
    )
    assert reflections.size() > 700
    assert reflections.size() < 1000



"""Contains classes used to construct a target function for refinement,
principally Target and ReflectionManager."""


from __future__ import annotations

import math
from typing import Any, Tuple, Union

from libtbx.phil import parse
from scitbx import sparse
from scitbx.array_family import flex

phil_str = """
    rmsd_cutoff = *fraction_of_bin_size absolute
      .help = "Method to choose rmsd cutoffs. This is currently either as a"
              "fraction of the discrete units of the spot positional data, i.e."
              "(pixel width, pixel height, image thickness in phi), or a tuple"
              "of absolute values to use as the cutoffs"
      .type = choice

    bin_size_fraction = 0.0
      .help = "Set this to a fractional value, say 0.2, to make a cut off in"
              "the natural discrete units of positional data, viz.,"
              "(pixel width, pixel height, image thickness in phi). This would"
              "then determine when the RMSD target is achieved. Only used if"
              "rmsd_cutoff = fraction_of_bin_size."
      .type = float(value_min=0.)

    absolute_cutoffs = None
      .help = "Absolute Values for the RMSD target achieved cutoffs in X, Y and"
              "Phi. The units are (mm, mm, rad)."
      .type = floats(size=3, value_min=0.)

    gradient_calculation_blocksize = None
      .help = "Maximum number of reflections to use for gradient calculation."
              "If there are more reflections than this in the manager then"
              "the minimiser must do the full calculation in blocks."
      .type = int(value_min=1)
"""
phil_scope = parse(phil_str)

# constants
RAD_TO_DEG = 180.0 / math.pi


class TargetFactory:
    @staticmethod
    def from_parameters_and_experiments(
        params,
        experiments,
        reflection_manager,
        predictor,
        pred_param,
        restraints_param,
        do_stills=False,
        do_sparse=False,
    ):
        """Given a set of parameters, configure a factory to build a
        target function

        Params:
            params The input parameters

        Returns:
            The target factory instance
        """

        if params.rmsd_cutoff == "fraction_of_bin_size":
            absolute_cutoffs = None
        elif params.rmsd_cutoff == "absolute":
            absolute_cutoffs = params.absolute_cutoffs
        else:
            raise RuntimeError(
                "Target function rmsd_cutoff option"
                + params.rmsd_cutoff
                + " not recognised"
            )

        # Determine whether the target is in X, Y, Phi space or just X, Y to choose
        # the right Target to instantiate
        if do_stills:
            if do_sparse:
                from dials.algorithms.refinement.target_stills import (
                    LeastSquaresStillsResidualWithRmsdCutoffSparse as targ,
                )
            else:
                from dials.algorithms.refinement.target_stills import (
                    LeastSquaresStillsResidualWithRmsdCutoff as targ,
                )
        else:
            if do_sparse:
                targ = LeastSquaresPositionalResidualWithRmsdCutoffSparse
            else:
                targ = LeastSquaresPositionalResidualWithRmsdCutoff

        # Here we pass in None for prediction_parameterisation and
        # restraints_parameterisation, as these are not required for initialisation
        # and will be linked to the object later
        return targ(
            experiments=experiments,
            predictor=predictor,
            reflection_manager=reflection_manager,
            prediction_parameterisation=pred_param,
            restraints_parameterisation=restraints_param,
            frac_binsize_cutoff=params.bin_size_fraction,
            absolute_cutoffs=absolute_cutoffs,
            gradient_calculation_blocksize=params.gradient_calculation_blocksize,
        )


class Target:
    """Abstract interface for a target function class

    A Target object will be used by a Refinery. It will refer to a Reflection
    Manager to get a list of observations. It will perform reflection prediction
    on those observations and update the reflection manager with those
    predictions. It will then query the reflection manager to get a list of
    accepted reflections with observed and calculated positions. These are the
    reflections for use in refinement. It obtains the gradients of reflection
    positions from a relevant prediction parameterisation object. With all of
    this information in place, it calculates the value of the target function,
    the gradients of the target function and auxiliary information (e.g. RMSDs).

    Concrete instances of this class implement the actual target function
    calculation. The base class should not determine what type of target
    function is used (e.g. least squares target), or limit whether the object
    is used for a detector space & phi residual, or a reciprocal space residual.
    This should all be set by a derived class.
    """

    _grad_names = ["dX_dp", "dY_dp", "dphi_dp"]
    rmsd_names = ["RMSD_X", "RMSD_Y", "RMSD_Phi"]
    rmsd_units = ["mm", "mm", "rad"]

    def __init__(
        self,
        experiments,
        predictor,
        reflection_manager,
        prediction_parameterisation,
        restraints_parameterisation=None,
        gradient_calculation_blocksize=None,
    ):

        self._experiments = experiments
        self._reflection_predictor = predictor
        self._reflection_manager = reflection_manager
        self._prediction_parameterisation = prediction_parameterisation
        self._restraints_parameterisation = restraints_parameterisation

        # Quantities to cache each step
        self._rmsds = None
        self._matches = None

        # Keep maximum number of reflections used for Jacobian calculation, if
        # a cutoff is required
        self._gradient_calculation_blocksize = gradient_calculation_blocksize

    @property
    def dim(self):
        """Get the number of dimensions of the target function"""
        return len(self._grad_names)

    def _predict_core(self, reflections, skip_derivatives=False):
        """perform prediction for the specified reflections"""

        # If the prediction parameterisation has a compose method (true for the scan
        # varying case) then call it. Prefer hasattr to try-except duck typing to
        # avoid masking AttributeErrors that could be raised within the method.
        if hasattr(self._prediction_parameterisation, "compose"):
            self._prediction_parameterisation.compose(reflections, skip_derivatives)

        # do prediction (updates reflection table in situ). Scan-varying prediction
        # is done automatically if the crystal has scan-points (assuming reflections
        # have ub_matrix set)
        self._reflection_predictor(reflections)

        x_obs, y_obs, phi_obs = reflections["xyzobs.mm.value"].parts()
        x_calc, y_calc, phi_calc = reflections["xyzcal.mm"].parts()

        # calculate residuals and assign columns
        reflections["x_resid"] = x_calc - x_obs
        reflections["x_resid2"] = reflections["x_resid"] ** 2
        reflections["y_resid"] = y_calc - y_obs
        reflections["y_resid2"] = reflections["y_resid"] ** 2
        reflections["phi_resid"] = phi_calc - phi_obs
        reflections["phi_resid2"] = reflections["phi_resid"] ** 2

        return reflections

    def predict(self):
        """perform reflection prediction for the working reflections and update the
        reflection manager"""

        # get the matches
        reflections = self._reflection_manager.get_obs()

        # reset the 'use' flag for all observations
        self._reflection_manager.reset_accepted_reflections()

        # predict
        reflections = self._predict_core(reflections)

        # set used_in_refinement flag to all those that had predictions
        mask = reflections.get_flags(reflections.flags.predicted)
        reflections.set_flags(mask, reflections.flags.used_in_refinement)

        # collect the matches
        self.update_matches(force=True)

    def predict_for_free_reflections(self):
        """perform prediction for the reflections not used for refinement"""

        refs = self._reflection_manager.get_free_reflections()
        if len(refs) == 0:
            return refs

        return self._predict_core(refs)

    def predict_for_reflection_table(self, reflections, skip_derivatives=False):
        """perform prediction for all reflections in the supplied table"""

        if "entering" not in reflections:
            reflections.calculate_entering_flags(self._experiments)

        # can only predict for experiments that exist and within the scan range
        # any other reflections will be left unchanged
        inc = flex.size_t_range(len(reflections))
        to_keep = flex.bool(len(inc), False)

        for iexp, exp in enumerate(self._experiments):
            sel = reflections["id"] == iexp

            # keep all reflections if there is no rotation axis
            if exp.goniometer is None:
                to_keep.set_selected(sel, True)
                continue

            # trim reflections outside the scan range
            phi = reflections["xyzobs.mm.value"].parts()[2]
            phi_min, phi_max = exp.scan.get_oscillation_range(deg=False)
            passed = (phi >= phi_min) & (phi <= phi_max)
            to_keep.set_selected(sel, passed)

        # determine indices to include and predict on the subset
        inc = inc.select(to_keep)
        sub_refl = reflections.select(inc)
        preds = self._predict_core(sub_refl, skip_derivatives)

        # set updated subset back into place
        reflections.set_selected(inc, preds)

        return reflections

    def calculate_gradients(self, reflections=None, callback=None):
        """delegate to the prediction_parameterisation object to calculate
        gradients for all the matched reflections, or just for those specified"""

        self.update_matches()

        if reflections:
            gradients = self._prediction_parameterisation.get_gradients(
                reflections, callback
            )
        else:
            gradients = self._prediction_parameterisation.get_gradients(
                self._matches, callback
            )

        return gradients

    def get_num_matches(self):
        """return the number of reflections currently used in the calculation"""

        self.update_matches()
        return len(self._matches)

    def get_num_matches_for_experiment(self, iexp=0):
        """return the number of reflections currently used in the calculation"""

        self.update_matches()
        sel = self._matches["id"] == iexp
        return sel.count(True)

    def get_num_matches_for_panel(self, ipanel=0):
        """return the number of reflections currently used in the calculation"""

        self.update_matches()
        sel = self._matches["panel"] == ipanel
        return sel.count(True)

    def update_matches(self, force=False):
        """ensure the observations matched to predictions are up to date"""

        if not self._matches or force:
            self._matches = self._reflection_manager.get_matches()

    def compute_functional_gradients_and_curvatures(self, block=None):
        """calculate the value of the target function and its gradients. Set
        approximate curvatures as a side-effect"""

        self.update_matches()
        if block is not None:
            matches = block
        else:
            matches = self._matches
        nref = len(matches)

        # This is a hack for the case where nref=0. This should not be necessary
        # if bounds are provided for parameters to stop the algorithm exploring
        # unreasonable regions of parameter space where no predictions exist.
        # Unfortunately the L-BFGS line search does make such extreme trials.
        if nref == 0:
            self._curv = [1.0] * len(self._prediction_parameterisation)
            return 1.0e12, [1.0] * len(self._prediction_parameterisation)

        residuals, weights = self._extract_residuals_and_weights(matches)
        w_resid = weights * residuals
        residuals2 = self._extract_squared_residuals(matches)

        # calculate target function
        L = 0.5 * flex.sum(weights * residuals2)

        def process_one_gradient(result):
            # copy gradients out of the result in the right order
            grads = [result[key] for key in self._grad_names]

            # reset result
            for k in result:
                result[k] = None

            # add new keys
            grads = self._concatenate_gradients(grads)
            result["dL_dp"] = flex.sum(w_resid * grads)
            result["curvature"] = flex.sum(weights * grads * grads)
            return result

        results = self.calculate_gradients(matches, callback=process_one_gradient)

        dL_dp = [result["dL_dp"] for result in results]
        curvs = [result["curvature"] for result in results]

        return (L, dL_dp, curvs)

    def compute_restraints_functional_gradients_and_curvatures(self):
        """use the restraints_parameterisation object, if present, to
        calculate the least squares restraints objective plus gradients and
        approximate curvatures"""

        if not self._restraints_parameterisation:
            return None

        (
            residuals,
            jacobian,
            weights,
        ) = self._restraints_parameterisation.get_residuals_gradients_and_weights()
        w_resid = weights * residuals
        residuals2 = residuals * residuals

        # calculate target function
        L = 0.5 * flex.sum(weights * residuals2)

        # calculate gradients using the scalar product of sparse vector col with
        # dense vector w_resid
        dL_dp = [col * w_resid for col in jacobian.cols()]

        # calculate lsq approximation to curvatures using weighted dot product
        curvs = [sparse.weighted_dot(col, weights, col) for col in jacobian.cols()]

        return (L, dL_dp, curvs)

    def compute_residuals(self):
        """return the vector of residuals plus their weights"""

        self.update_matches()
        return self._extract_residuals_and_weights(self._matches)

    def split_matches_into_blocks(self, nproc: int = 1):
        """Return a list of the matches, split into blocks according to the
        gradient_calculation_blocksize parameter and the number of processes (if relevant).
        The number of blocks will be set such that the total number of reflections
        being processed by concurrent processes does not exceed gradient_calculation_blocksize"""

        self.update_matches()

        # Need to be able to track the indices of the original matches table for
        # scan-varying gradient calculations. A simple and robust (but slightly
        # expensive) way to do this is to add an index column to the matches table
        self._matches["imatch"] = flex.size_t_range(len(self._matches))

        if self._gradient_calculation_blocksize:
            nblocks = int(
                math.floor(
                    len(self._matches) * nproc / self._gradient_calculation_blocksize
                )
            )
        else:
            nblocks = nproc
        # ensure at least 100 reflections per block
        nblocks = min(nblocks, int(len(self._matches) / 100))
        nblocks = max(nblocks, 1)
        blocksize = int(math.floor(len(self._matches) / nblocks))
        blocks = []
        for block_num in range(nblocks - 1):
            start = block_num * blocksize
            end = (block_num + 1) * blocksize
            blocks.append(self._matches[start:end])
        start = (nblocks - 1) * blocksize
        end = len(self._matches)
        blocks.append(self._matches[start:end])
        return blocks

    def compute_residuals_and_gradients(
        self, block=None
    ) -> Tuple[Any, Union[flex.double, sparse.matrix], Any]:
        """return the vector of residuals plus their gradients and weights for
        non-linear least squares methods"""

        self.update_matches()
        if block is not None:
            matches = block
        else:
            matches = self._matches

        # The gradients are provided as a list of dictionaries, where the
        # dictionaries have keys corresponding to gradients of the different types
        # of residual involved. For example, for scans the keys are 'dX_dp',
        # 'dY_dp', 'dphi_dp'. Reshape this data structure so that all the gradients
        # of a particular type of residual are kept together
        gradients = self.calculate_gradients(matches)
        reshaped = []
        for key in self._grad_names:
            reshaped.append([g[key] for g in gradients])

        residuals, weights = self._extract_residuals_and_weights(matches)

        nelem = len(matches) * len(self._grad_names)
        nparam = len(self._prediction_parameterisation)
        jacobian = self._build_jacobian(reshaped, nelem=nelem, nparam=nparam)

        return (residuals, jacobian, weights)

    def compute_restraints_residuals_and_gradients(self):
        """delegate to the restraints_parameterisation object, if present, to
        calculate the vector of restraints residuals plus their gradients and
        weights for non-linear least squares methods"""

        if self._restraints_parameterisation:
            (
                residuals,
                jacobian,
                weights,
            ) = self._restraints_parameterisation.get_residuals_gradients_and_weights()
            return (residuals, jacobian, weights)

        else:
            return None

    @staticmethod
    def _build_jacobian(
        grads_each_dim, nelem=None, nparam=None
    ) -> Union[sparse.matrix, flex.double]:
        """construct Jacobian from lists of gradient vectors. The elements of
        grads_each_dim refer to the gradients of each dimension of the problem
        (e.g. dX, dY, dZ). The elements for a single dimension give the arrays
        of gradients of the associated residual for each parameter. This method
        may be overridden for the case where these vectors use sparse storage"""

        jacobian = flex.double(flex.grid(nelem, nparam))
        # loop over parameters
        for i in range(nparam):
            col = grads_each_dim[0][i]
            for g in grads_each_dim[1:]:
                col.extend(g[i])
            jacobian.matrix_paste_column_in_place(col, i)

        return jacobian

    @staticmethod
    def _concatenate_gradients(grads):
        """concatenate gradient vectors and return a flex.double. This method
        may be overridden for the case where these vectors use sparse storage"""

        result = grads[0]
        for g in grads[1:]:
            result.extend(g)
        return result

    @staticmethod
    def _extract_residuals_and_weights(matches):
        """extract vector of residuals and corresponding weights. The space the
        residuals are measured in (e.g. X, Y and Phi) and the order they are
        returned is determined by a concrete implementation of a staticmethod"""
        raise NotImplementedError()

    @staticmethod
    def _extract_squared_residuals(matches):
        """extract vector of squared residuals. The space the residuals are measured
        in (e.g. X, Y and Phi) and the order they are returned is determined by a
        concrete implementation of a staticmethod"""
        raise NotImplementedError()

    def rmsds(self):
        """calculate unweighted RMSDs for the matches"""

        self.update_matches()

        # cache rmsd calculation for achieved test
        self._rmsds = self._rmsds_core(self._matches)

        return self._rmsds

    def rmsds_for_reflection_table(self, reflections):
        """calculate unweighted RMSDs for the specified reflections. Caution: this
        assumes that the table reflections has the keys expected by _rmsds_core"""

        n = len(reflections)
        if n == 0:
            return None
        return self._rmsds_core(reflections)

    def rmsds_for_experiment(self, iexp=0):
        """calculate unweighted RMSDs for the selected experiment."""

        self.update_matches()
        sel = self._matches["id"] == iexp
        n = sel.count(True)
        if n == 0:
            return None

        rmsds = self._rmsds_core(self._matches.select(sel))
        return rmsds

    def rmsds_for_panel(self, ipanel=0):
        """calculate unweighted RMSDs for the selected panel."""

        self.update_matches()
        sel = self._matches["panel"] == ipanel
        n = sel.count(True)
        if n == 0:
            return None

        rmsds = self._rmsds_core(self._matches.select(sel))
        return rmsds

    def achieved(self):
        """return True to terminate the refinement."""
        raise NotImplementedError()


class LeastSquaresPositionalResidualWithRmsdCutoff(Target):
    """An implementation of the target class providing a least squares residual
    in terms of detector impact position X, Y and phi, terminating on achieved
    rmsd (or on intrisic convergence of the chosen minimiser)"""

    def __init__(
        self,
        experiments,
        predictor,
        reflection_manager,
        prediction_parameterisation,
        restraints_parameterisation,
        frac_binsize_cutoff=0.33333,
        absolute_cutoffs=None,
        gradient_calculation_blocksize=None,
    ):

        Target.__init__(
            self,
            experiments,
            predictor,
            reflection_manager,
            prediction_parameterisation,
            restraints_parameterisation,
            gradient_calculation_blocksize,
        )

        # Set up the RMSD achieved criterion. For simplicity, we take models from
        # the first Experiment only. If this is not appropriate for refinement over
        # all experiments then absolute cutoffs should be used instead.
        if experiments[0].scan:
            image_width_rad = abs(experiments[0].scan.get_oscillation(deg=False)[1])
        else:
            image_width_rad = None
        detector = experiments[0].detector

        if not absolute_cutoffs:
            pixel_sizes = [p.get_pixel_size() for p in detector]
            min_px_size_x = min(e[0] for e in pixel_sizes)
            min_px_size_y = min(e[1] for e in pixel_sizes)
            self._binsize_cutoffs = [
                min_px_size_x * frac_binsize_cutoff,
                min_px_size_y * frac_binsize_cutoff,
                image_width_rad * frac_binsize_cutoff,
            ]
        else:
            assert len(absolute_cutoffs) == 3
            self._binsize_cutoffs = absolute_cutoffs

    @staticmethod
    def _extract_residuals_and_weights(matches) -> Tuple[flex.double, Any]:

        # return residuals and weights as 1d flex.double vectors
        residuals = flex.double.concatenate(matches["x_resid"], matches["y_resid"])
        residuals.extend(matches["phi_resid"])

        weights, w_y, w_z = matches["xyzobs.mm.weights"].parts()
        weights.extend(w_y)
        weights.extend(w_z)

        return residuals, weights

    @staticmethod
    def _extract_squared_residuals(matches):

        residuals2 = flex.double.concatenate(matches["x_resid2"], matches["y_resid2"])
        residuals2.extend(matches["phi_resid2"])

        return residuals2

    def _rmsds_core(self, reflections):
        """calculate unweighted RMSDs for the specified reflections"""

        resid_x = flex.sum(reflections["x_resid2"])
        resid_y = flex.sum(reflections["y_resid2"])
        resid_phi = flex.sum(reflections["phi_resid2"])
        n = len(reflections)

        rmsds = (
            math.sqrt(resid_x / n),
            math.sqrt(resid_y / n),
            math.sqrt(resid_phi / n),
        )
        return rmsds

    def achieved(self):
        """RMSD criterion for target achieved"""
        r = self._rmsds if self._rmsds else self.rmsds()

        # reset cached rmsds to avoid getting out of step
        self._rmsds = None

        if (
            r[0] < self._binsize_cutoffs[0]
            and r[1] < self._binsize_cutoffs[1]
            and r[2] < self._binsize_cutoffs[2]
        ):
            return True
        return False


class SparseGradientsMixin:
    """Mixin class to build a sparse Jacobian from gradients of the prediction
    formula stored as sparse vectors, and allow concatenation of gradient vectors
    that employed sparse storage."""

    @staticmethod
    def _build_jacobian(grads_each_dim, nelem=None, nparam=None):
        """construct Jacobian from lists of sparse gradient vectors."""

        nref = int(nelem / len(grads_each_dim))

        blocks = [sparse.matrix(nref, nparam) for _ in grads_each_dim]
        jacobian = sparse.matrix(nelem, nparam)

        # loop over parameters, building full width blocks of the full Jacobian
        for i in range(nparam):
            for j, block in enumerate(blocks):
                grad = grads_each_dim[j]
                block[:, i] = grad[i]
                # Having copied this block, delete the source data
                grads_each_dim[j][i] = None

        # set the blocks in the Jacobian
        for i, block in enumerate(blocks):
            jacobian.assign_block(block, (i * nref), 0)
            blocks[i] = None

        return jacobian

    @staticmethod
    def _concatenate_gradients(grads):
        """concatenate sparse gradient vectors and return a flex.double."""

        result = grads[0].as_dense_vector()
        for g in grads[1:]:
            result.extend(g.as_dense_vector())
        return result


class LeastSquaresPositionalResidualWithRmsdCutoffSparse(
    SparseGradientsMixin, LeastSquaresPositionalResidualWithRmsdCutoff
):
    """A version of the LeastSquaresPositionalResidualWithRmsdCutoff Target that
    uses a sparse matrix data structure for memory efficiency when there are a
    large number of Experiments"""

    pass



from __future__ import annotations

import json

import procrunner
import pytest

from dials.command_line import estimate_resolution as cmdline


@pytest.mark.parametrize(
    "input_files",
    [
        ("AUTOMATIC_DEFAULT_scaled_unmerged.mtz",),
        ("AUTOMATIC_DEFAULT_scaled.refl", "AUTOMATIC_DEFAULT_scaled.expt"),
    ],
)
def test_x4wide(input_files, dials_data, run_in_tmp_path, capsys):
    x4wide = dials_data("x4wide_processed", pathlib=True)
    paths = [str(x4wide / p) for p in input_files]
    reference_mtz = x4wide / "AUTOMATIC_DEFAULT_scaled.mtz"
    result = cmdline.run(
        [
            "cc_half=0.9",
            "isigma=2",
            "misigma=3",
            "rmerge=0.5",
            "completeness=1.0",
            "i_mean_over_sigma_mean=3",
            "batch_range=1,20",
            "batch_range=70,90",
            "space_group=P43212",
            f"reference={reference_mtz}",
            "cc_ref=0.9",
            "labels=IMEAN,SIGIMEAN",
            "html=resolutionizer.html",
            "json=resolutionizer.json",
        ]
        + paths,
    )
    captured = capsys.readouterr()
    expected_output = (
        "Resolution rmerge:        1.34",
        "Resolution cc_half:       1.56",
        "Resolution cc_ref:        1.3",
        "Resolution I/sig:         1.53",
        "Resolution Mn(I/sig):     1.51",
        "Resolution Mn(I)/Mn(sig): 1.50",
    )
    for line in expected_output:
        assert line in captured.out
    assert run_in_tmp_path.joinpath("resolutionizer.html").is_file()
    expected_keys = {
        "cc_half",
        "cc_ref",
        "isigma",
        "misigma",
        "i_mean_over_sigma_mean",
        "rmerge",
        "completeness",
    }
    assert set(result.keys()) == expected_keys
    resolutionizer = run_in_tmp_path / "resolutionizer.json"
    assert resolutionizer.is_file()
    with resolutionizer.open() as fh:
        d = json.load(fh)
    assert set(d.keys()) == expected_keys


def test_multi_sequence_with_batch_range(dials_data, run_in_tmp_path, capsys):
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refls = location / "scaled_20_25.refl"
    expts = location / "scaled_20_25.expt"

    cmdline.run(
        ["batch_range=1900,3600", str(refls), str(expts)],
    )
    captured = capsys.readouterr()

    expected_output = "Resolution cc_half:       0.61"
    for line in expected_output:
        assert line in captured.out
    assert run_in_tmp_path.joinpath("dials.estimate_resolution.html").is_file()


def test_dispatcher_name():
    result = procrunner.run(["dials.estimate_resolution"])
    assert not result.returncode
    assert not result.stderr


def test_handle_fit_failure(dials_data, run_in_tmp_path, capsys):
    location = dials_data("l_cysteine_dials_output", pathlib=True)
    cmdline.run(
        ["misigma=1"]
        + [
            str(location / f)
            for f in (
                "11_integrated.expt",
                "11_integrated.refl",
                "23_integrated.expt",
                "23_integrated.refl",
            )
        ]
    )
    captured = capsys.readouterr()

    expected_output = (
        "Resolution fit against cc_half failed: No reflections left for fitting",
        "Resolution Mn(I/sig):     0.62",
    )
    for line in expected_output:
        assert line in captured.out
    assert run_in_tmp_path.joinpath("dials.estimate_resolution.html").is_file()


def test_mismatched_experiments_reflections(dials_data, run_in_tmp_path):
    location = dials_data("l_cysteine_dials_output", pathlib=True)
    with pytest.raises(SystemExit):
        cmdline.run(
            [
                str(location / f)
                for f in (
                    "11_integrated.expt",
                    "11_integrated.refl",
                    "23_integrated.refl",
                )
            ]
        )



from __future__ import annotations

import os

from libtbx import easy_run


def test_compare_orientation_matrices(dials_regression, run_in_tmp_path):
    path = os.path.join(dials_regression, "refinement_test_data", "i04_weak_data")
    cmd = (
        "dials.compare_orientation_matrices %s/experiments.json %s/regression_experiments.json"
        % (path, path)
    )
    result = easy_run.fully_buffered(cmd).raise_if_errors()
    out = "\n".join(result.stdout_lines[7:])
    out = out.replace("-0", "0")
    assert (
        out
        == """\
Change of basis op: a,b,c
Rotation matrix to transform crystal 1 to crystal 2:
{{1.000, 0.000, 0.000},
 {0.000, 1.000, 0.000},
 {0.000, 0.000, 1.000}}
Rotation of 0.002 degrees about axis (0.916, 0.081, 0.393)
"""
    )



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1


from __future__ import annotations

import pickle
import sys

import iotbx.phil

import dials.util.log
from dials.util.image_viewer.spotfinder_wrap import spot_wrapper
from dials.util.options import ArgumentParser, flatten_experiments, flatten_reflections

help_message = """

This program can be used for viewing diffraction images, optionally overlaid
with the results of spot finding, indexing or integration.

Examples::

  dials.image_viewer image.cbf

  dials.image_viewer models.expt

  dials.image_viewer models.expt strong.refl

  dials.image_viewer models.expt integrated.refl
"""

phil_scope = iotbx.phil.parse(
    """\
brightness = 10
  .type = int
color_scheme = *grayscale rainbow heatmap invert
  .type = choice
projection = lab *image
  .type = choice
show_beam_center = True
  .type = bool
show_resolution_rings = False
  .type = bool
show_ice_rings = False
  .type = bool
show_ctr_mass = True
  .type = bool
show_max_pix = True
  .type = bool
show_all_pix = True
  .type = bool
show_threshold_pix = False
  .type = bool
show_shoebox = True
  .type = bool
show_predictions = True
  .type = bool
show_miller_indices = False
  .type = bool
show_indexed = False
  .type = bool
show_integrated = False
  .type = bool
show_mask = False
  .type = bool
show_basis_vectors = True
  .type = bool
show_rotation_axis = False
  .type = bool
basis_vector_scale = 10
  .type = int(value_min=1, value_max=20)
display = *image mean variance dispersion sigma_b \
          sigma_s threshold global_threshold
  .type = choice
nsigma_b = 6
  .type = float(value_min=0)
nsigma_s = 3
  .type = float(value_min=0)
global_threshold = 0
  .type = float(value_min=0)
kernel_size = 3,3
  .type = ints(size=2, value_min=1)
min_local = 2
  .type = int
gain = 1
  .type = float(value_min=0)
  .help = "Set gain for the dispersion algorithm. This does not override the"
          "detector's panel gain, but acts as a multiplier for it."

include scope dials.extensions.radial_profile_spotfinder_threshold_ext.phil_str

stack_images = 1
  .type = int(value_min=1)
  .expert_level = 2
stack_mode = max mean *sum
  .type = choice
d_min = None
  .type = float(value_min=0)
mask = None
  .type = path
  .help = path to mask pickle file

powder_arcs{
  show = False
    .type=bool
    .help = "show powder arcs calculated from PDB file."
  code = None
    .type=str
    .help = "PDB code (4 characters) for file; fetch it from the Internet."
}
calibrate_silver = False
    .type=bool
    .help = "Open special GUI for distance/metrology from silver behenate."
calibrate_pdb{
  code = None
    .type=str
    .help = "Open pdb code (over Internet) to get unit cell & symmetry for powder rings."
    .help = "Most useful for calibrating low-Q rings on far detector."
    .help = "Option is mutually exclusive with calibrate silver, unit cell and powder arcs options."
  d_min = 20.
    .type=float
    .help = "Limiting resolution to calculate powder rings"
}
calibrate_unit_cell{
  unit_cell = None
    .type=unit_cell
    .help = "Specify unit cell for powder rings."
    .help = "Option is mutually exclusive with calibrate silver, pdb and powder arcs options."
  d_min = 20.
    .type=float
    .help = "Limiting resolution to calculate powder rings"
  space_group = None
    .type=str
    .help = "Specify spacegroup for the unit cell"
  show_hkl = None
    .type = ints(size=3)
    .multiple=True
    .help = "Limit display of rings to these Miller indices"
}

include scope dials.util.options.format_phil_scope

masking {
  include scope dials.util.masking.phil_scope
}

output {
  mask = pixels.mask
    .type = path
    .help = "Name of output mask file"

  mask_params = mask.phil
    .type = path
    .help = "Name of output mask parameter file"
}

predict_reflections = False
  .type = bool
  .help = Predict reflections if no reflections provided in input

include scope dials.algorithms.profile_model.factory.phil_scope
include scope dials.algorithms.spot_prediction.reflection_predictor.phil_scope

load_models = True
  .type = bool
  .help = "Whether to load every model, which matters for large image files"

zmq_endpoint = None
  .type = str
  .help = "The endpoint to bind a zeromq PULL socket to, for receiving commands"
  .expert_level = 3
""",
    process_includes=True,
)


def show_image_viewer(params, experiments, reflections):
    wrapper = spot_wrapper(params=params)
    wrapper.display(experiments=experiments, reflections=reflections)


@dials.util.show_mail_handle_errors()
def run(args=None):
    dials.util.log.print_banner()
    usage_message = "dials.image_viewer models.expt [observations.refl]"
    parser = ArgumentParser(
        usage=usage_message,
        phil=phil_scope,
        read_experiments=True,
        read_reflections=True,
        read_experiments_from_images=True,
        epilog=help_message,
    )
    params, options = parser.parse_args(args, show_diff_phil=True)
    experiments = [x.data for x in params.input.experiments]
    reflections = flatten_reflections(params.input.reflections)

    if len(experiments) == 0:
        parser.print_help()
        exit(0)

    flat_expts = flatten_experiments(params.input.experiments)
    if params.load_models:
        if any(e.detector is None for e in flat_expts):
            sys.exit("Error: experiment has no detector")
        if any(e.beam is None for e in flat_expts):
            sys.exit("Error: experiment has no beam")

    # If given a mask, replace the path with the loaded data
    if params.mask is not None:
        with open(params.mask, "rb") as f:
            params.mask = pickle.load(f)

    show_image_viewer(params=params, reflections=reflections, experiments=experiments)


if __name__ == "__main__":
    run()



from __future__ import annotations

import random

from dials.array_family import flex
from dials.util.reciprocal_lattice import Render3d


def test_Render3d(mocker, multi_sequence_data):
    experiments = multi_sequence_data["experiments"]
    reflections = multi_sequence_data["reflections"]
    render = Render3d()
    render.viewer = mocker.Mock()
    mocker.spy(render, "set_beam_centre")
    mocker.spy(render, "map_points_to_reciprocal_space")
    mocker.spy(render, "set_points")
    render.load_models(experiments, reflections)
    assert render.set_beam_centre.call_count == 0
    assert render.map_points_to_reciprocal_space.call_count == 1
    assert render.set_points.call_count == 1
    assert render.viewer.set_points.call_count == 1
    assert render.viewer.set_points.call_args[0][0].size() == 1255
    assert render.viewer.set_colors.call_count == 1
    assert render.viewer.set_palette.call_count == 1
    assert render.viewer.set_reciprocal_lattice_vectors.call_count == 1
    assert list(render.viewer.set_palette.call_args[0][0][:2]) == [
        (1.0, 1.0, 1.0),
        (0.9019607843137255, 0.6235294117647059, 0.0),
    ]

    render.settings.beam_centre = (211, 205)
    render.settings.beam_centre = (41, 97)
    render.load_models(experiments, reflections)
    assert render.set_beam_centre.call_count == 1

    for (outlier_display, expected_count) in (("outliers", 0), ("inliers", 1255)):
        render.settings.outlier_display = outlier_display
        render.load_models(experiments, reflections)
        assert render.viewer.set_points.call_args[0][0].size() == expected_count

    for (display, expected_count) in (
        ("indexed", 1255),
        ("unindexed", 0),
        ("integrated", 0),
    ):
        render.settings.display = display
        render.load_models(experiments, reflections)
        assert render.viewer.set_points.call_args[0][0].size() == expected_count

    render.settings.display = "all"
    render.settings.experiment_ids = [0, 2, 3]
    render.load_models(experiments, reflections)
    assert render.viewer.set_points.call_args[0][0].size() == 957

    render.settings.black_background = False
    render.load_models(experiments, reflections)
    assert list(render.viewer.set_palette.call_args[0][0][:2]) == [
        (0.0, 0.0, 0.0),
        (0.0980392156862745, 0.3764705882352941, 1.0),
    ]

    render.settings.reverse_phi = True
    mocker.spy(flex.reflection_table, "map_centroids_to_reciprocal_space")
    render.load_models(experiments, reflections)
    assert flex.reflection_table.map_centroids_to_reciprocal_space.call_count == 1


def test_Render3d_integrated(mocker, centroid_test_data):
    experiments = centroid_test_data["experiments"]
    reflections = centroid_test_data["reflections"]

    render = Render3d()
    render.viewer = mocker.Mock()
    render.load_models(experiments, reflections)
    assert render.viewer.set_points.call_args[0][0].size() == 2269

    render.settings.partiality_min = 0.2
    render.settings.partiality_max = 0.9
    render.load_models(experiments, reflections)
    assert render.viewer.set_points.call_args[0][0].size() == 923

    render = Render3d()
    render.viewer = mocker.Mock()
    render.settings.z_min = 2
    render.settings.z_max = 7
    render.load_models(experiments, reflections)
    assert render.viewer.set_points.call_args[0][0].size() == 1885

    random.seed(42)
    reflections["n_signal"] = flex.size_t(
        random.randint(1, 100) for i in range(len(reflections))
    )

    render = Render3d()
    render.viewer = mocker.Mock()
    render.settings.n_min = None
    render.settings.n_max = None
    render.load_models(experiments, reflections)
    assert render.viewer.set_points.call_args[0][0].size() == 2269

    render = Render3d()
    render.viewer = mocker.Mock()
    render.settings.n_min = 20
    render.settings.n_max = 80
    render.load_models(experiments, reflections)
    expected = 1368
    assert render.viewer.set_points.call_args[0][0].size() == expected



from __future__ import annotations

from time import strftime

import h5py

from dials.util.nexus import nx_mx, nx_reflections


def get_entry(filename, mode="a"):
    handle = h5py.File(filename, mode)
    if "entry" in handle:
        entry = handle["entry"]
        assert entry.attrs["NX_class"] == "NXentry"
    else:
        entry = handle.create_group("entry")
        entry.attrs["NX_class"] = "NXentry"
        handle.attrs["file_name"] = filename
        handle.attrs["file_time"] = strftime("%Y-%m-%dT%H:%M:%S")
    return entry


def load(filename):
    entry = get_entry(filename, "r")
    ref, exp_index = nx_reflections.load(entry)
    exp = nx_mx.load(entry, exp_index)
    return exp, ref


def dump(experiments, reflections, params):
    filename = params.hklout
    entry = get_entry(filename, "w")
    experiments = nx_mx.dump(entry, experiments, params)
    nx_reflections.dump(entry, reflections, experiments)



from __future__ import annotations

import math
import random

import pytest

from dxtbx.serialize import load
from scitbx import matrix

from dials.algorithms.profile_model.gaussian_rs import (
    BBoxCalculator3D,
    CoordinateSystem,
)


@pytest.fixture
def setup(dials_data):
    sequence = load.imageset(
        dials_data("centroid_test_data", pathlib=True) / "sweep.json"
    )

    fixture = {}

    # Get the models
    fixture["beam"] = sequence.get_beam()
    fixture["detector"] = sequence.get_detector()
    fixture["gonio"] = sequence.get_goniometer()
    fixture["scan"] = sequence.get_scan()

    # Set the delta_divergence/mosaicity
    n_sigma = 5
    sigma_divergence = 0.060 * math.pi / 180
    mosaicity = 0.154 * math.pi / 180
    fixture["delta_divergence"] = n_sigma * sigma_divergence
    fixture["delta_mosaicity"] = n_sigma * mosaicity

    # Create the bounding box calculator
    fixture["calculate_bbox"] = BBoxCalculator3D(
        fixture["beam"],
        fixture["detector"],
        fixture["gonio"],
        fixture["scan"],
        fixture["delta_divergence"],
        fixture["delta_mosaicity"],
    )
    return fixture


def test_outer_bounds(setup):
    assert len(setup["detector"]) == 1
    s0 = setup["beam"].get_s0()
    m2 = setup["gonio"].get_rotation_axis()
    s0_length = matrix.col(setup["beam"].get_s0()).length()
    width, height = setup["detector"][0].get_image_size()
    zrange = setup["scan"].get_array_range()

    for i in range(1000):
        # Get random x, y, z
        x = random.uniform(0, 2000)
        y = random.uniform(0, 2000)
        z = random.uniform(0, 9)

        # Get random s1, phi, panel
        s1 = (
            matrix.col(setup["detector"][0].get_pixel_lab_coord((x, y))).normalize()
            * s0_length
        )
        phi = setup["scan"].get_angle_from_array_index(z, deg=False)
        panel = 0

        # Calculate the bounding box
        bbox = setup["calculate_bbox"](s1, z, panel)
        z1, z2 = bbox[4], bbox[5]

        # Calculate the rotation angle for each point
        phi_dash1 = setup["scan"].get_angle_from_array_index(z1, deg=False)
        phi_dash2 = setup["scan"].get_angle_from_array_index(z2, deg=False)

        # Create the XDS coordinate system
        xcs = CoordinateSystem(m2, s0, s1, phi)

        # Calculate reciprocal space coordinates at each point
        e11, e21, e31 = xcs.from_beam_vector_and_rotation_angle(s1, phi_dash1)
        e12, e22, e32 = xcs.from_beam_vector_and_rotation_angle(s1, phi_dash2)

        # Check vertical edges
        for j in range(bbox[2], bbox[3] + 1):
            xyz1 = setup["detector"][0].get_pixel_lab_coord((bbox[0], j))
            xyz2 = setup["detector"][0].get_pixel_lab_coord((bbox[1] + 1, j))
            sdash1 = matrix.col(xyz1).normalize() * s0_length
            sdash2 = matrix.col(xyz2).normalize() * s0_length
            e11, e21, e3 = xcs.from_beam_vector_and_rotation_angle(sdash1, phi)
            e12, e22, e3 = xcs.from_beam_vector_and_rotation_angle(sdash2, phi)
            if bbox[0] > 0 and bbox[1] < width:
                assert (
                    abs(e11) >= setup["delta_divergence"]
                    or abs(e21) >= setup["delta_divergence"]
                )
                assert (
                    abs(e12) >= setup["delta_divergence"]
                    or abs(e22) >= setup["delta_divergence"]
                )

        # Check horizontal edges
        for i in range(bbox[0], bbox[1] + 1):
            xyz1 = setup["detector"][0].get_pixel_lab_coord((i, bbox[2]))
            xyz2 = setup["detector"][0].get_pixel_lab_coord((i, bbox[3] + 1))
            sdash1 = matrix.col(xyz1).normalize() * s0_length
            sdash2 = matrix.col(xyz2).normalize() * s0_length
            e11, e21, e3 = xcs.from_beam_vector_and_rotation_angle(sdash1, phi)
            e12, e22, e3 = xcs.from_beam_vector_and_rotation_angle(sdash2, phi)
            if bbox[2] > 0 and bbox[3] < height:
                assert (
                    abs(e11) >= setup["delta_divergence"]
                    or abs(e21) >= setup["delta_divergence"]
                )
                assert (
                    abs(e12) >= setup["delta_divergence"]
                    or abs(e22) >= setup["delta_divergence"]
                )

        # All e3 coords >= delta_mosaicity
        if bbox[4] > zrange[0] and bbox[5] < zrange[1]:
            assert abs(e31) >= setup["delta_mosaicity"]
            assert abs(e32) >= setup["delta_mosaicity"]


def test_radius(setup):
    s0 = setup["beam"].get_s0()
    m2 = setup["gonio"].get_rotation_axis()
    s0_length = matrix.col(setup["beam"].get_s0()).length()

    width, height = setup["detector"][0].get_image_size()
    radius12 = setup["delta_divergence"]

    for i in range(1000):
        # Get random x, y, z
        x = random.uniform(0, 2000)
        y = random.uniform(0, 2000)
        z = random.uniform(0, 9)

        # Get random s1, phi, panel
        s1 = (
            matrix.col(setup["detector"][0].get_pixel_lab_coord((x, y))).normalize()
            * s0_length
        )
        phi = setup["scan"].get_angle_from_array_index(z, deg=False)
        panel = 0

        # Calculate the bounding box
        bbox = setup["calculate_bbox"](s1, z, panel)

        # Create the XDS coordinate system
        xcs = CoordinateSystem(m2, s0, s1, phi)

        # Check vertical edges
        for j in range(bbox[2], bbox[3] + 1):
            xyz1 = setup["detector"][0].get_pixel_lab_coord((bbox[0], j))
            xyz2 = setup["detector"][0].get_pixel_lab_coord((bbox[1] + 1, j))
            sdash1 = matrix.col(xyz1).normalize() * s0_length
            sdash2 = matrix.col(xyz2).normalize() * s0_length
            e11, e21, e31 = xcs.from_beam_vector_and_rotation_angle(sdash1, phi)
            e12, e22, e31 = xcs.from_beam_vector_and_rotation_angle(sdash2, phi)
            if bbox[0] > 0 and bbox[1] < width:
                assert math.sqrt(e11**2 + e21**2) >= radius12
                assert math.sqrt(e12**2 + e22**2) >= radius12

        # Check horizontal edges
        for i in range(bbox[0], bbox[1] + 1):
            xyz1 = setup["detector"][0].get_pixel_lab_coord((i, bbox[2]))
            xyz2 = setup["detector"][0].get_pixel_lab_coord((i, bbox[3] + 1))
            sdash1 = matrix.col(xyz1).normalize() * s0_length
            sdash2 = matrix.col(xyz2).normalize() * s0_length
            e11, e21, e32 = xcs.from_beam_vector_and_rotation_angle(sdash1, phi)
            e12, e22, e32 = xcs.from_beam_vector_and_rotation_angle(sdash2, phi)
            if bbox[2] > 0 and bbox[3] < height:
                assert math.sqrt(e11**2 + e21**2) >= radius12
                assert math.sqrt(e12**2 + e22**2) >= radius12



from __future__ import annotations

import logging

import libtbx
from scitbx import matrix
from scitbx.array_family import flex

from dials.algorithms.spot_finding.threshold import DispersionExtendedThresholdStrategy

logger = logging.getLogger(
    "dials.extensions.dispersion_extended_spotfinder_threshold_ext"
)


class DispersionExtendedSpotFinderThresholdExt:
    """Extensions to do dispersion threshold."""

    name = "dispersion_extended"

    default = True

    @staticmethod
    def phil():
        return None

    def __init__(self, params):
        """
        Initialise the algorithm.

        :param params: The input parameters
        """
        self.params = params

    def compute_threshold(self, image, mask, **kwargs):
        r"""
        Compute the threshold.

        :param image: The image to process
        :param mask: The pixel mask on the image
        :\*\*kwargs: Arbitrary keyword arguments
        :returns: A boolean mask showing foreground/background pixels
        """

        params = self.params
        if params.spotfinder.threshold.dispersion.global_threshold is libtbx.Auto:
            params.spotfinder.threshold.dispersion.global_threshold = int(
                estimate_global_threshold(image, mask)
            )
            logger.info(
                "Setting global_threshold: %i",
                params.spotfinder.threshold.dispersion.global_threshold,
            )

        self._algorithm = DispersionExtendedThresholdStrategy(
            kernel_size=params.spotfinder.threshold.dispersion.kernel_size,
            gain=params.spotfinder.threshold.dispersion.gain,
            mask=params.spotfinder.lookup.mask,
            n_sigma_b=params.spotfinder.threshold.dispersion.sigma_background,
            n_sigma_s=params.spotfinder.threshold.dispersion.sigma_strong,
            min_count=params.spotfinder.threshold.dispersion.min_local,
            global_threshold=params.spotfinder.threshold.dispersion.global_threshold,
        )

        return self._algorithm(image, mask)


def estimate_global_threshold(image, mask=None, plot=False):
    n_above_threshold = flex.size_t()
    threshold = flex.double()
    for i in range(1, 20):
        g = 1.5**i
        g = int(g)
        n_above_threshold.append((image > g).count(True))
        threshold.append(g)

    # Find the elbow point of the curve, in the same manner as that used by
    # distl spotfinder for resolution method 1 (Zhang et al 2006).
    # See also dials/algorithms/spot_finding/per_image_analysis.py

    x = threshold.as_double()
    y = n_above_threshold.as_double()
    slopes = (y[-1] - y[:-1]) / (x[-1] - x[:-1])
    p_m = flex.min_index(slopes)

    x1 = matrix.col((x[p_m], y[p_m]))
    x2 = matrix.col((x[-1], y[-1]))

    gaps = flex.double()
    v = matrix.col(((x2[1] - x1[1]), -(x2[0] - x1[0]))).normalize()

    for i in range(p_m, len(x)):
        x0 = matrix.col((x[i], y[i]))
        r = x1 - x0
        g = abs(v.dot(r))
        gaps.append(g)

    p_g = flex.max_index(gaps)

    x_g_ = x[p_g + p_m]
    y_g_ = y[p_g + p_m]

    # more conservative, choose point 2 left of the elbow point
    x_g = x[p_g + p_m - 2]
    # y_g = y[p_g + p_m - 2]

    if plot:
        from matplotlib import pyplot

        pyplot.figure(figsize=(16, 12))
        pyplot.scatter(threshold, n_above_threshold, marker="+")
        # for i in range(len(threshold)-1):
        #  pyplot.plot([threshold[i], threshold[-1]],
        #              [n_above_threshold[i], n_above_threshold[-1]])
        # for i in range(1, len(threshold)):
        #  pyplot.plot([threshold[0], threshold[i]],
        #              [n_above_threshold[0], n_above_threshold[i]])
        pyplot.plot([x_g, x_g], pyplot.ylim())
        pyplot.plot(
            [threshold[p_m], threshold[-1]],
            [n_above_threshold[p_m], n_above_threshold[-1]],
        )
        pyplot.plot([x_g_, threshold[-1]], [y_g_, n_above_threshold[-1]])
        pyplot.xlabel("Threshold")
        pyplot.ylabel("Number of pixels above threshold")
        pyplot.savefig("global_threshold.png")
        pyplot.clf()

    return x_g



from __future__ import annotations

import copy
import math
import os

import pytest


def generate_reflections(experiments):
    from cctbx.sgtbx import space_group, space_group_symbols
    from scitbx.array_family import flex

    from dials.algorithms.refinement.prediction.managed_predictors import (
        ScansExperimentsPredictor,
        ScansRayPredictor,
    )
    from dials.algorithms.spot_prediction import IndexGenerator, ray_intersection

    detector = experiments[0].detector
    crystal = experiments[0].crystal

    # All indices in a 2.0 Angstrom sphere
    resolution = 2.0
    index_generator = IndexGenerator(
        crystal.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices = index_generator.to_array()

    # Predict rays within the sequence range
    scan = experiments[0].scan
    sequence_range = scan.get_oscillation_range(deg=False)
    ray_predictor = ScansRayPredictor(experiments, sequence_range)
    obs_refs = ray_predictor(indices)

    # Take only those rays that intersect the detector
    intersects = ray_intersection(detector, obs_refs)
    obs_refs = obs_refs.select(intersects)

    # Make a reflection predictor and re-predict for all these reflections. The
    # result is the same, but we gain also the flags and xyzcal.px columns
    ref_predictor = ScansExperimentsPredictor(experiments)
    obs_refs["id"] = flex.int(len(obs_refs), 0)
    obs_refs = ref_predictor(obs_refs)

    # Set 'observed' centroids from the predicted ones
    obs_refs["xyzobs.mm.value"] = obs_refs["xyzcal.mm"]

    # Invent some variances for the centroid positions of the simulated data
    im_width = 0.1 * math.pi / 180.0
    px_size = detector[0].get_pixel_size()
    var_x = flex.double(len(obs_refs), (px_size[0] / 2.0) ** 2)
    var_y = flex.double(len(obs_refs), (px_size[1] / 2.0) ** 2)
    var_phi = flex.double(len(obs_refs), (im_width / 2.0) ** 2)
    obs_refs["xyzobs.mm.variance"] = flex.vec3_double(var_x, var_y, var_phi)

    return obs_refs, ref_predictor


def test1(dials_regression):
    # use a experiments that contains a CS-PAD detector description
    data_dir = os.path.join(dials_regression, "refinement_test_data", "hierarchy_test")
    experiments_path = os.path.join(data_dir, "datablock.json")
    assert os.path.exists(experiments_path)

    # load models
    from dxtbx.model.experiment_list import ExperimentListFactory

    experiments = ExperimentListFactory.from_serialized_format(
        experiments_path, check_format=False
    )
    im_set = experiments.imagesets()[0]
    detector = copy.deepcopy(im_set.get_detector())
    beam = im_set.get_beam()

    # we'll invent a crystal, goniometer and scan for this test
    from dxtbx.model import Crystal

    crystal = Crystal(
        (40.0, 0.0, 0.0), (0.0, 40.0, 0.0), (0.0, 0.0, 40.0), space_group_symbol="P1"
    )

    from dxtbx.model import GoniometerFactory

    goniometer = GoniometerFactory.known_axis((1.0, 0.0, 0.0))

    # Build a mock scan for a 180 degree sequence
    from dxtbx.model import ScanFactory

    sf = ScanFactory()
    scan = sf.make_scan(
        image_range=(1, 1800),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(1800)),
        deg=True,
    )
    sequence_range = scan.get_oscillation_range(deg=False)
    im_width = scan.get_oscillation(deg=False)[1]
    assert sequence_range == (0.0, math.pi)
    assert im_width == pytest.approx(0.1 * math.pi / 180.0)

    from dxtbx.model.experiment_list import Experiment, ExperimentList

    # Build an experiment list
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            crystal=crystal,
            imageset=None,
        )
    )

    # simulate some reflections
    refs, ref_predictor = generate_reflections(experiments)

    # move the detector quadrants apart by 2mm both horizontally and vertically
    from dials.algorithms.refinement.parameterisation.detector_parameters import (
        DetectorParameterisationHierarchical,
    )

    det_param = DetectorParameterisationHierarchical(detector, level=1)
    det_p_vals = det_param.get_param_vals()
    p_vals = list(det_p_vals)
    p_vals[1] += 2
    p_vals[2] -= 2
    p_vals[7] += 2
    p_vals[8] += 2
    p_vals[13] -= 2
    p_vals[14] += 2
    p_vals[19] -= 2
    p_vals[20] -= 2
    det_param.set_param_vals(p_vals)

    # reparameterise the detector at the new perturbed geometry
    det_param = DetectorParameterisationHierarchical(detector, level=1)

    # parameterise other models
    from dials.algorithms.refinement.parameterisation.beam_parameters import (
        BeamParameterisation,
    )
    from dials.algorithms.refinement.parameterisation.crystal_parameters import (
        CrystalOrientationParameterisation,
        CrystalUnitCellParameterisation,
    )

    beam_param = BeamParameterisation(beam, goniometer)
    xlo_param = CrystalOrientationParameterisation(crystal)
    xluc_param = CrystalUnitCellParameterisation(crystal)

    # fix beam
    beam_param.set_fixed([True] * 3)

    # fix crystal
    xluc_param.set_fixed([True] * 6)
    xlo_param.set_fixed([True] * 3)

    # parameterisation of the prediction equation
    from dials.algorithms.refinement.parameterisation.parameter_report import (
        ParameterReporter,
    )
    from dials.algorithms.refinement.parameterisation.prediction_parameters import (
        XYPhiPredictionParameterisation,
    )

    pred_param = XYPhiPredictionParameterisation(
        experiments, [det_param], [beam_param], [xlo_param], [xluc_param]
    )
    param_reporter = ParameterReporter(
        [det_param], [beam_param], [xlo_param], [xluc_param]
    )

    # reflection manager and target function
    from dials.algorithms.refinement.reflection_manager import ReflectionManager
    from dials.algorithms.refinement.target import (
        LeastSquaresPositionalResidualWithRmsdCutoff,
    )

    refman = ReflectionManager(refs, experiments, nref_per_degree=20)

    # set a very tight rmsd target of 1/10000 of a pixel
    target = LeastSquaresPositionalResidualWithRmsdCutoff(
        experiments,
        ref_predictor,
        refman,
        pred_param,
        restraints_parameterisation=None,
        frac_binsize_cutoff=0.0001,
    )

    # minimisation engine
    from dials.algorithms.refinement.engine import (
        LevenbergMarquardtIterations as Refinery,
    )

    refinery = Refinery(
        target=target,
        prediction_parameterisation=pred_param,
        log=None,
        max_iterations=20,
    )

    # Refiner
    from dials.algorithms.refinement.refiner import Refiner

    refiner = Refiner(
        experiments=experiments,
        pred_param=pred_param,
        param_reporter=param_reporter,
        refman=refman,
        target=target,
        refinery=refinery,
    )

    history = refiner.run()
    assert history.reason_for_termination == "RMSD target achieved"

    # compare detector with original detector
    orig_det = im_set.get_detector()
    refined_det = refiner.get_experiments()[0].detector

    from scitbx import matrix

    for op, rp in zip(orig_det, refined_det):
        # compare the origin vectors by...
        o1 = matrix.col(op.get_origin())
        o2 = matrix.col(rp.get_origin())
        # ...their relative lengths
        assert math.fabs(o1.length() - o2.length()) / o1.length() == pytest.approx(
            0, abs=1e-5
        )
        # ...the angle between them
        assert o1.accute_angle(o2) == pytest.approx(0, abs=1e-5)



from __future__ import annotations

from unittest.mock import patch

import pytest

pytest.importorskip("pyFAI")

from dxtbx.serialize import load

from dials.command_line import powder_calibrate
from dials.command_line.powder_calibrate import Geometry, Point, PowderCalibrator


@pytest.mark.parametrize(
    "eyeball, starting_geometry", [(True, "imported.expt"), (False, "eyeballed.expt")]
)
def test_calibrate_coarse(dials_data, tmp_path, eyeball, starting_geometry):

    aluminium_powder = dials_data("aluminium_standard", pathlib=True)

    starting_geom_exptlist = load.experiment_list(aluminium_powder / starting_geometry)

    def mocked_eyeball(self):
        """
        Mock the calibrate method to update obj geometry to an eyeballed one
        without calling matplotlib Widget tools
        """
        self.geometry.update_beam_pos(
            beam_coords_px=Point(1103, 1024), beam_dist_mm=635
        )

    with patch.object(
        powder_calibrate.EyeballWidget,
        "eyeball",
        new=mocked_eyeball,
    ):
        test_calibrator = PowderCalibrator(
            expts=starting_geom_exptlist,
            standard="Al",
            calibrated_geometry=str(tmp_path / "test_calibrated.expt"),
            pyfai_improvement=str(tmp_path / "test_pyfai_improvement.png"),
            straight_lines=str(tmp_path / "test_straight_lines.png"),
        )
        if eyeball:
            test_calibrator.calibrate_with_widget()
        test_calibrator.refine_with_pyfai(plots=False)

    calibrated_geom = test_calibrator.geometry

    expected_geom_exptlist = load.experiment_list(aluminium_powder / "calibrated.expt")
    expected_geom = Geometry(expt=expected_geom_exptlist)

    assert pytest.approx(calibrated_geom.param, 1e-3) == expected_geom.param


def test_save_geom_to_expt(dials_data, tmp_path):

    aluminium_powder = dials_data("aluminium_standard", pathlib=True)

    imported_exptlist = load.experiment_list(aluminium_powder / "imported.expt")
    imported_geom = Geometry(expt=imported_exptlist)

    outfile = tmp_path / "test_save.expt"
    imported_geom.save_to_expt(output=outfile)
    assert outfile.is_file()

    from_saved_expt = load.experiment_list(outfile)
    read_from_saved_geom = Geometry(expt=from_saved_expt)

    assert imported_geom.param == pytest.approx(read_from_saved_geom.param)



from __future__ import annotations

import sys
import threading

import wx

from .slip_viewer.frame import chooser_wrapper as _chooser_wrapper
from .viewer_tools import ZeroMQEvent

try:
    import platform
    import resource

    def debug_memory_usage():
        # getrusage returns kb on linux, bytes on mac
        units_per_mb = 1024
        if platform.system() == "Darwin":
            units_per_mb = 1024 * 1024
        print(
            "Memory usage: %.1f MB"
            % (int(resource.getrusage(resource.RUSAGE_SELF).ru_maxrss) / units_per_mb)
        )

except ImportError:

    def debug_memory_usage():
        pass


class chooser_wrapper(_chooser_wrapper):
    def show_header(self):
        #   debug_memory_usage()
        pass

    def __eq__(self, other):
        return (
            hasattr(other, "image_set")
            and self.image_set is other.image_set
            and hasattr(other, "index")
            and self.index == other.index
        )

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((id(self.image_set), self.index))


class spot_wrapper:
    def __init__(self, params):
        self.params = params
        self.frame = None
        self._zmq_thread = None

    def display(self, experiments, reflections):
        import wx

        from dials.util.image_viewer.spotfinder_frame import SpotFrame

        app = wx.App()

        self.frame = SpotFrame(
            None,
            -1,
            "X-ray image display",
            size=(800, 720),
            pos=(100, 100),
            params=self.params,
            experiments=experiments,
            reflections=reflections,
        )
        self.frame.SetSize((1024, 780))
        self.frame.Show()

        imagesets = self.frame.imagesets
        for imageset in imagesets:
            for idx in range(len(imageset.indices())):
                self.frame.add_file_name_or_data(chooser_wrapper(imageset, idx))
        # Make sure we load the first image as the current one
        self.frame.load_image(chooser_wrapper(imagesets[0], 0))

        # If ZMQ required, setup the endpoint to dispatch messages to wx
        if self.params.zmq_endpoint is not None:
            self._setup_zmq_endpoint(self.params.zmq_endpoint)

        #   debug_memory_usage()
        app.MainLoop()

        # Shut down the ZMQ endpoint if we created one
        if self.params.zmq_endpoint is not None:
            self._shutdown_zmq_endpoint()

    def load_image(self, filename):
        from dials.util.spotfinder_frame import create_load_image_event

        if self.frame is not None:
            create_load_image_event(self.frame, filename)

    def _setup_zmq_endpoint(self, endpoint: str) -> None:
        """Create and bind a ZMQ endpoint"""
        try:
            import zmq
        except ImportError:
            sys.exit("Error: ZMQ Endpoint requested but pyzmq not installed; aborting")
        print("ZeroMQ binding requested at", endpoint)
        self.zmq_context = zmq.Context()
        self.zmq_socket = self.zmq_context.socket(zmq.PULL)
        self.zmq_socket.bind(endpoint)
        self._zmq_thread = ZMQEventListener(self.zmq_socket, self.frame)
        self._zmq_thread.start()

    def _shutdown_zmq_endpoint(self):
        """Close the ZMQ endpoint"""
        print("Shutting down ZMQ endpoint")
        # Wait for the thread to end
        if self._zmq_thread is not None:
            self._zmq_thread.stop = True
            self._zmq_thread.join()

        self.zmq_socket.close()


class ZMQEventListener(threading.Thread):
    """Thread to listen for ZMQ events and dispatch them to the GUI.

    Args:
        socket: The ZeroMQ socket object to listen on
        target: The WX container to forward messages to
    """

    def __init__(self, socket, target, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stop = False
        self.socket = socket
        self.target = target

    def run(self):
        while not self.stop:
            event_count = self.socket.poll(timeout=0.1)
            for _ in range(event_count):
                message = self.socket.recv_json()
                print("Received ZMQ Message: ", message)
                self._handle_message(message)

    def _handle_message(self, message):
        # Just attach the message to an event and send it on
        evt = ZeroMQEvent(message=message)
        wx.PostEvent(self.target, evt)



"""
Observers for the cosym procedure.
"""

from __future__ import annotations

import json

from jinja2 import ChoiceLoader, Environment, PackageLoader

from dials.algorithms.clustering.observers import UnitCellAnalysisObserver
from dials.algorithms.symmetry.cosym import SymmetryAnalysis
from dials.algorithms.symmetry.cosym.plots import plot_coords, plot_rij_histogram
from dials.util.observer import Observer, singleton


def register_default_cosym_observers(script):
    """Register the standard observers to the cosym script."""
    script.cosym_analysis.register_observer(
        event="analysed_symmetry", observer=SymmetryAnalysisObserver()
    )
    script.cosym_analysis.register_observer(
        event="optimised", observer=CosymClusterAnalysisObserver()
    )
    script.register_observer(event="run_cosym", observer=UnitCellAnalysisObserver())
    script.register_observer(
        event="run_cosym", observer=CosymHTMLGenerator(), callback="make_html"
    )
    script.register_observer(
        event="run_cosym", observer=CosymJSONGenerator(), callback="make_json"
    )


@singleton
class CosymHTMLGenerator(Observer):
    """
    Observer to make a html report
    """

    def make_html(self, cosym_script):
        """Collect data from the individual observers and write the html."""
        filename = cosym_script.params.output.html
        if not filename:
            return
        self.data.update(CosymClusterAnalysisObserver().make_plots())
        self.data.update(UnitCellAnalysisObserver().make_plots())
        self.data.update(SymmetryAnalysisObserver().make_tables())
        print(f"Writing html report to: {filename}")
        loader = ChoiceLoader(
            [
                PackageLoader("dials", "templates"),
                PackageLoader("dials", "static", encoding="utf-8"),
            ]
        )
        env = Environment(loader=loader)
        template = env.get_template("cosym_report.html")
        html = template.render(
            page_title="DIALS cosym report",
            cosym_graphs=self.data["cosym_graphs"],
            unit_cell_graphs=self.data["unit_cell_graphs"],
            symmetry_analysis=self.data["symmetry_analysis"],
        )
        with open(filename, "wb") as f:
            f.write(html.encode("utf-8", "xmlcharrefreplace"))


@singleton
class CosymJSONGenerator(Observer):
    """
    Observer to make a html report
    """

    def make_json(self, cosym_script):
        """Collect data from the individual observers and write the html."""
        filename = cosym_script.params.output.json
        if not filename:
            return
        self.data.update(CosymClusterAnalysisObserver().make_plots())
        self.data.update(UnitCellAnalysisObserver().make_plots())
        self.data.update(SymmetryAnalysisObserver().get_data())
        print(f"Writing json to: {filename}")
        with open(filename, "w") as f:
            json.dump(self.data, f)


@singleton
class CosymClusterAnalysisObserver(Observer):
    """
    Observer to record cosym cluster analysis data and make model plots.
    """

    def update(self, cosym):
        """Update the data in the observer."""
        self.data["coordinates"] = cosym.coords
        self.data["rij_matrix"] = cosym.target.rij_matrix

    def make_plots(self):
        """Generate cosym cluster analysis plot data."""
        d = plot_rij_histogram(self.data["rij_matrix"])
        d.update(plot_coords(self.data["coordinates"]))
        graphs = {"cosym_graphs": d}
        return graphs


@singleton
class SymmetryAnalysisObserver(Observer):
    """
    Observer to record symmetry analysis data and make tables.
    """

    def update(self, cosym):
        if cosym._symmetry_analysis is not None:
            self.data.update(cosym._symmetry_analysis.as_dict())

    def make_tables(self):
        """Generate symmetry analysis tables."""
        d = {"symmetry_analysis": {}}
        if self.data:
            d["symmetry_analysis"].update(
                {
                    "sym_ops_table": SymmetryAnalysis.sym_ops_table(self.data),
                    "subgroups_table": SymmetryAnalysis.subgroups_table(self.data),
                    "summary_table": SymmetryAnalysis.summary_table(self.data),
                }
            )
        return d

    def get_data(self):
        return self.data



from __future__ import annotations

import numpy as np


def plot_coords(coords, labels=None, key="cosym_coordinates"):

    coord_x = coords[:, 0]
    coord_y = coords[:, 1]
    assert coord_x.size == coord_y.size, (coord_x.size, coord_y.size)

    if labels is None:
        labels = np.full(coord_x.shape[0], -1, dtype=int)

    unique_labels = set(labels)
    unique_labels = sorted(unique_labels)
    n_clusters = max(len(unique_labels) - (1 if -1 in unique_labels else 0), 1)

    # XXX should avoid relying on matplotlib here to determine colours
    import matplotlib

    matplotlib.use("Agg")
    from matplotlib import pyplot as plt

    colours = plt.cm.Spectral(np.linspace(0, 1, n_clusters)).tolist()

    if -1 in unique_labels:
        colours.insert(0, (0, 0, 0, 1))
    data = []
    for k, col in zip(unique_labels, colours):
        isel = np.where(labels == k)[0]
        data.append(
            {
                "x": coord_x[isel].tolist(),
                "y": coord_y[isel].tolist(),
                "mode": "markers",
                "type": "scatter",
                "marker": {
                    "size": 2,
                    "alpha": 0.5,
                    "color": "rgb(%f,%f,%f)" % tuple(col[:3]),
                },
                "name": "Cluster %i" % k if k >= 0 else "Noise",
            }
        )
    d = {
        key: {
            "data": data,
            "layout": {
                "title": "Cosym coordinates",
                "xaxis": {"range": [-1, 1], "constrain": "domain"},
                "yaxis": {"range": [-1, 1], "scaleanchor": "x", "constrain": "domain"},
            },
            "help": """\
The outcome of the dials.cosym multi-dimensional scaling procedure projected on to two
dimensions. Each point corresponds to an individual data set, or symmetry copy of a data
set. The lengths of the vectors are inversely related to the amount of random error in
each data set, and can be interpreted as an estimate of the CC* values. The angular
separation between any pair, or groups, of vectors is a measure of the systematic
differences between the data sets, for example as a result of an indexing ambiguity,
or the presence of non-isomorphism.
""",
        }
    }

    return d


def plot_rij_histogram(rij_matrix, key="cosym_rij_histogram"):
    """Plot a histogram of the rij values.

    Args:
      plot_name (str): The file name to save the plot to.
        If this is not defined then the plot is displayed in interactive mode.
    """
    hist, bin_edges = np.histogram(
        rij_matrix[rij_matrix != 0],
        bins=100,
        range=(min(-1, rij_matrix.min()), max(1, rij_matrix.max())),
    )
    bin_centers = bin_edges[:-1] + np.diff(bin_edges) / 2

    d = {
        key: {
            "data": [
                {
                    "x": bin_centers.tolist(),
                    "y": hist.tolist(),
                    "type": "bar",
                    "name": "Rij histogram",
                }
            ],
            "layout": {
                "title": "Distribution of values in the Rij matrix",
                "xaxis": {"title": "r<sub>ij</sub>"},
                "yaxis": {"title": "Frequency"},
                "bargap": 0,
            },
            "help": """\
A histogram of the values of the Rij matrix of pairwise correlation coefficients. A
unimodal distribution of values may suggest that no indexing ambiguity is evident,
whereas a bimodal distribution can be indicative of the presence of an indexing
ambiguity.
""",
        }
    }
    return d



from __future__ import annotations

from pathlib import Path

import procrunner
import pytest

from dxtbx.model.experiment_list import ExperimentListFactory


def plot_beam_centre_error(ideal_bc, obs_bc):
    import matplotlib.pyplot as plt

    ideal_x, ideal_y = zip(*ideal_bc)
    obs_x, obs_y = zip(*obs_bc)
    del_x = [a - b for a, b in zip(obs_x, ideal_x)]
    del_y = [a - b for a, b in zip(obs_y, ideal_y)]
    scan_points = range(len(ideal_x))
    plt.plot(scan_points, del_x, scan_points, del_y)
    plt.xlabel("Scan point")
    plt.ylabel("Beam centre residual (obs - ideal) (pixels)")
    plt.show()


def test_refinement_and_compare_with_known_truth(dials_regression, tmp_path):
    # use data generated by simulation for this test
    data_dir = (
        Path(dials_regression) / "refinement_test_data" / "varying_beam_direction"
    )
    experiments_path = data_dir / "refined_static.json"
    pickle_path = data_dir / "refined_static.pickle"

    assert experiments_path.is_file()
    assert pickle_path.is_file()

    # Run refinement and load models
    result = procrunner.run(
        [
            "dials.refine",
            experiments_path,
            pickle_path,
            "scan_varying=True",
            "crystal.orientation.force_static=True",
            "crystal.unit_cell.force_static=True",
            "beam.force_static=False",
            "beam.fix=wavelength",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    exp = ExperimentListFactory.from_json_file(
        tmp_path / "refined.expt", check_format=False
    )[0]
    beam, detector = exp.beam, exp.detector

    # Beam centre at every scan-point
    s0_scan_points = [beam.get_s0_at_scan_point(i) for i in range(beam.num_scan_points)]
    bc_scan_points = [detector[0].get_beam_centre_px(s0) for s0 in s0_scan_points]

    # Set up the nanoBragg object as used in the simulation
    from .sim_images import Simulation

    sim = Simulation()
    sim.set_varying_beam(along="both")

    # Simulation beam centre at every scan-point
    sim_s0_scan_points = [
        sim.beam.get_s0_at_scan_point(i) for i in range(sim.beam.num_scan_points)
    ]
    sim_bc_scan_points = [
        sim.detector[0].get_beam_centre_px(s0) for s0 in sim_s0_scan_points
    ]

    assert beam.num_scan_points == sim.beam.num_scan_points

    # Generate a plot. This shows that the beam centre is worse at the edges of
    # the scan. This is what we expect as the centroids at the scan edges are
    # least well determined because of the truncation of found spots. At these
    # edges the error approaches 0.15 pixels, whilst in the central region of
    # the scan it is within 0.05 pixels.
    #
    # plot_beam_centre_error(sim_bc_scan_points, bc_scan_points)

    # Compare the results.
    for bc1, bc2 in zip(sim_bc_scan_points, bc_scan_points):
        assert bc2 == pytest.approx(bc1, abs=0.15)



from __future__ import annotations


def test_run(dials_regression, run_in_tmp_path):
    from os.path import join

    from dxtbx.model.experiment_list import ExperimentListFactory

    from dials.array_family import flex
    from dials.command_line.export import phil_scope
    from dials.util.nexus import dump, load

    path = join(dials_regression, "nexus_test_data")

    # Read the experiments
    experiments1 = ExperimentListFactory.from_json_file(
        join(path, "refined_experiments.json")
    )

    # Read the reflections
    reflections1 = flex.reflection_table.from_file(join(path, "integrated.pickle"))

    # Delete some columns for the test
    del reflections1["s1"]
    del reflections1["zeta"]
    del reflections1["background.mse"]

    # Dump the reflections
    params = phil_scope.extract()
    params.nxs.hklout = "hklout.nxs"
    dump(experiments1, reflections1, params.nxs)

    # Load them again
    experiments2, reflections2 = load("hklout.nxs")

    EPS = 1e-7

    # Check the reflections are OK
    assert reflections1.nrows() == reflections2.nrows()
    assert reflections1.ncols() == reflections2.ncols()
    for key in reflections1.keys():
        data1 = reflections1[key]
        data2 = reflections2[key]
        assert data1.__class__ == data2.__class__
        if isinstance(data1, flex.double):
            assert data1.all_approx_equal(data2)
        elif isinstance(data1, flex.int6):
            for p1, p2 in zip(data1.parts(), data2.parts()):
                assert p1.all_eq(p2)
        elif isinstance(data1, flex.vec3_double):
            for p1, p2 in zip(data1.parts(), data2.parts()):
                assert p1.all_approx_equal(p2)
        else:
            assert data1.all_eq(data2)

    # Test passed

    # Check the experiments are ok
    assert len(experiments1) == len(experiments2)
    exp1 = experiments1[0]
    exp2 = experiments2[0]

    # Check the beam
    b1 = exp1.beam
    b2 = exp2.beam
    assert all(
        abs(d1 - d2) < EPS
        for d1, d2 in zip(
            b1.get_sample_to_source_direction(), b2.get_sample_to_source_direction()
        )
    )
    assert abs(b1.get_wavelength() - b2.get_wavelength()) < EPS
    assert abs(b1.get_polarization_fraction() - b2.get_polarization_fraction()) < EPS
    assert all(
        abs(d1 - d2) < EPS
        for d1, d2 in zip(b1.get_polarization_normal(), b2.get_polarization_normal())
    )

    # Check the goniometer
    g1 = exp1.goniometer
    g2 = exp2.goniometer
    assert all(
        abs(d1 - d2) < EPS
        for d1, d2 in zip(g1.get_rotation_axis(), g2.get_rotation_axis())
    )
    assert all(
        abs(d1 - d2) < EPS
        for d1, d2 in zip(g1.get_fixed_rotation(), g2.get_fixed_rotation())
    )
    assert all(
        abs(d1 - d2) < EPS
        for d1, d2 in zip(g1.get_setting_rotation(), g2.get_setting_rotation())
    )

    # Check the scan
    s1 = exp1.scan
    s2 = exp2.scan
    assert len(s1) == len(s2)
    assert s1.get_image_range() == s2.get_image_range()
    assert abs(s1.get_oscillation()[0] - s1.get_oscillation()[0]) < EPS
    assert abs(s1.get_oscillation()[1] - s1.get_oscillation()[1]) < EPS
    for e1, e2 in zip(s1.get_exposure_times(), s2.get_exposure_times()):
        assert abs(e1 - e2) < EPS
    for e1, e2 in zip(s1.get_epochs(), s2.get_epochs()):
        assert abs(e1 - e2) < EPS

    # Check the detector
    d1 = exp1.detector
    d2 = exp2.detector
    assert len(d1) == len(d2)
    for p1, p2 in zip(d1, d2):
        assert p1.get_type() == p2.get_type()
        assert p1.get_material() == p2.get_material()
        assert p1.get_thickness() == p2.get_thickness()
        assert p1.get_image_size() == p2.get_image_size()
        assert p1.get_pixel_size() == p2.get_pixel_size()
        assert p1.get_trusted_range() == p2.get_trusted_range()
        for x1, x2 in zip(p1.get_fast_axis(), p2.get_fast_axis()):
            assert abs(x1 - x2) < EPS
        for x1, x2 in zip(p1.get_slow_axis(), p2.get_slow_axis()):
            assert abs(x1 - x2) < EPS
        for x1, x2 in zip(p1.get_origin(), p2.get_origin()):
            assert abs(x1 - x2) < EPS

    # Check the crystal
    c1 = exp1.crystal
    c2 = exp2.crystal
    assert c1.get_space_group() == c2.get_space_group()
    for p1, p2 in zip(c1.get_unit_cell().parameters(), c2.get_unit_cell().parameters()):
        assert abs(p1 - p2) < EPS
    for p1, p2 in zip(c1.get_A(), c2.get_A()):
        assert abs(p1 - p2) < EPS
    assert c1.num_scan_points == c2.num_scan_points
    for i in range(c1.num_scan_points):
        A1 = c1.get_A_at_scan_point(i)
        A2 = c2.get_A_at_scan_point(i)
        for a1, a2 in zip(A1, A2):
            assert abs(a1 - a2) < EPS
        uc1 = c1.get_unit_cell_at_scan_point(i)
        uc2 = c2.get_unit_cell_at_scan_point(i)
        for p1, p2 in zip(uc1.parameters(), uc2.parameters()):
            assert abs(p1 - p2) < EPS



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Dict, List, Optional

import numpy as np
from numpy.linalg import norm

from dxtbx import flumpy
from scitbx import linalg, matrix

from dials.algorithms.profile_model.ellipsoid import mosaicity_from_eigen_decomposition
from dials.algorithms.refinement.parameterisation.crystal_parameters import (
    CrystalOrientationParameterisation,
    CrystalUnitCellParameterisation,
)
from dials.array_family import flex


class BaseParameterisation(ABC):
    def __init__(self, params: Optional[np.array] = None) -> None:
        """
        Initialise with the parameters

        """
        if params is not None:
            assert len(params) == self.num_parameters()
            self.params = params
        else:
            self.params = np.array([0.0] * self.num_parameters(), dtype=np.float64)

    @abstractmethod
    def num_parameters(self):
        pass

    @property
    def parameters(self) -> np.array:
        """
        Return the parameters

        """
        return self.params

    @parameters.setter
    def parameters(self, params: np.array) -> None:
        assert len(params) == self.num_parameters()
        self.params = params


class Simple1MosaicityParameterisation(BaseParameterisation):
    """
    A simple mosaicity parameterisation that uses 1 parameter to describe a
    multivariate normal reciprocal lattice profile. Sigma is enforced as positive
    definite by parameterising using the cholesky decomposition.

    M = | b1 0  0  |
        |  0 b1 0  |
        |  0  0 b1 |

    S = M*M^T

    """

    @staticmethod
    def is_angular() -> bool:
        return False

    @staticmethod
    def num_parameters() -> int:
        return 1

    def sigma(self) -> np.array:
        """
        Compute the covariance matrix of the MVN from the parameters

        """
        psq = self.params[0] ** 2
        return np.array(
            [[psq, 0, 0], [0, psq, 0], [0, 0, psq]],
            dtype=np.float64,
        )

    def first_derivatives(self) -> np.array:
        """
        Compute the first derivatives of Sigma w.r.t the parameters

        """
        b1 = self.params[0]
        return np.array(
            [[[2.0 * b1, 0, 0], [0, 2.0 * b1, 0], [0, 0, 2.0 * b1]]], dtype=np.float64
        ).reshape(1, 3, 3)

    def mosaicity(self) -> Dict:
        """One value for mosaicity for Simple1"""
        decomp = linalg.eigensystem.real_symmetric(
            matrix.sqr(flumpy.from_numpy(self.sigma())).as_flex_double_matrix()
        )
        v = list(mosaicity_from_eigen_decomposition(decomp.values()))
        return {"spherical": v[0]}


class Simple6MosaicityParameterisation(BaseParameterisation):
    """
    A simple mosaicity parameterisation that uses 6 parameters to describe a
    multivariate normal reciprocal lattice profile. Sigma is enforced as positive
    definite by parameterising using the cholesky decomposition.

    M = | b1 0  0  |
        | b2 b3 0  |
        | b4 b5 b6 |

    S = M*M^T

    """

    @staticmethod
    def is_angular() -> bool:
        return False

    @staticmethod
    def num_parameters() -> int:
        return 6

    def sigma(self) -> np.array:
        """
        Compute the covariance matrix of the MVN from the parameters

        """
        M = np.array(
            [
                [self.params[0], 0, 0],
                [self.params[1], self.params[2], 0],
                [self.params[3], self.params[4], self.params[5]],
            ],
            dtype=np.float64,
        )
        return np.matmul(M, M.T)

    def mosaicity(self) -> Dict:
        """Three components for mosaicity for Simple6"""
        decomp = linalg.eigensystem.real_symmetric(
            matrix.sqr(flumpy.from_numpy(self.sigma())).as_flex_double_matrix()
        )
        vals = list(mosaicity_from_eigen_decomposition(decomp.values()))
        min_m = min(vals)
        max_m = max(vals)
        vals.remove(min_m)
        vals.remove(max_m)
        return {"min": min_m, "mid": vals[0], "max": max_m}

    def first_derivatives(self) -> np.array:
        """
        Compute the first derivatives of Sigma w.r.t the parameters

        """
        b1, b2, b3, b4, b5, b6 = self.params

        d1 = np.array(
            [[2 * b1, b2, b4], [b2, 0, 0], [b4, 0, 0]], dtype=np.float64
        ).reshape(1, 3, 3)

        d2 = np.array(
            [[0, b1, 0], [b1, 2 * b2, b4], [0, b4, 0]], dtype=np.float64
        ).reshape(1, 3, 3)

        d3 = np.array(
            [[0, 0, 0], [0, 2 * b3, b5], [0, b5, 0]], dtype=np.float64
        ).reshape(1, 3, 3)

        d4 = np.array(
            [[0, 0, b1], [0, 0, b2], [b1, b2, 2 * b4]], dtype=np.float64
        ).reshape(1, 3, 3)

        d5 = np.array(
            [[0, 0, 0], [0, 0, b3], [0, b3, 2 * b5]], dtype=np.float64
        ).reshape(1, 3, 3)

        d6 = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 2 * b6]], dtype=np.float64).reshape(
            1, 3, 3
        )

        ds = np.concatenate([d1, d2, d3, d4, d5, d6], axis=0)

        return ds


class WavelengthSpreadParameterisation(BaseParameterisation):
    """
    A simple wavelength parameterisation that uses 1 parameter to describe a
    multivariate normal wavelength spread. Sigma is enforced as positive
    definite by parameterising using the cholesky decomposition.

    L = | 0 0 0  |
        | 0 0 0  |
        | 0 0 l1 |

    S = L*L^T

    """

    @staticmethod
    def num_parameters() -> int:
        return 1

    def sigma(self) -> flex.double:
        """
        The normal distribution sigma

        """
        return flex.double([self.params[0]])

    def first_derivatives(self) -> flex.double:
        """
        Compute the first derivatives of Sigma w.r.t the parameters

        """
        return flex.double([2 * self.params[0]])


class Angular2MosaicityParameterisation(BaseParameterisation):
    """
    A simple mosaicity parameterisation that uses 2 parameters to describe a
    multivariate normal angular mosaic spread. Sigma is enforced as positive
    definite by parameterising using the cholesky decomposition.
    W = | w1 0  0  |
        | 0 w1  0  |
        | 0  0 w2 |
    S = W*W^T
    """

    @staticmethod
    def is_angular() -> bool:
        return True

    @staticmethod
    def num_parameters() -> int:
        return 2

    def sigma(self) -> matrix:
        """
        Compute the covariance matrix of the MVN from the parameters
        """
        p1sq = self.params[0] ** 2
        p2sq = self.params[1] ** 2
        return np.array(
            [[p1sq, 0, 0], [0, p1sq, 0], [0, 0, p2sq]],
            dtype=np.float64,
        )

    def mosaicity(self) -> Dict:
        """Two unique components of mosaicity"""
        decomp = linalg.eigensystem.real_symmetric(
            matrix.sqr(flumpy.from_numpy(self.sigma())).as_flex_double_matrix()
        )
        m = mosaicity_from_eigen_decomposition(decomp.values())
        v = decomp.vectors()
        mosaicities = {"radial": 0, "angular": 0}
        # two values must be same, could have accidental degeneracy where all 3 same:
        unique_ = list(set(m))
        if len(unique_) == 1:
            return {"angular": unique_[0], "radial": unique_[0]}
        else:
            assert len(unique_) == 2
            for i in range(3):
                vec = (v[i * 3], v[(i * 3) + 1], v[(i * 3) + 2])
                if vec == (0, 0, 1):
                    mosaicities["radial"] = m[i]
                else:
                    mosaicities["angular"] = m[i]
        return mosaicities

    def first_derivatives(self) -> np.array:
        """
        Compute the first derivatives of Sigma w.r.t the parameters
        """
        b1, b2 = self.params

        d1 = np.array(
            [[2 * b1, 0, 0], [0, 2 * b1, 0], [0, 0, 0]], dtype=np.float64
        ).reshape(1, 3, 3)

        d2 = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 2 * b2]], dtype=np.float64).reshape(
            1, 3, 3
        )

        return np.concatenate([d1, d2], axis=0)


class Angular4MosaicityParameterisation(BaseParameterisation):
    """
    A simple mosaicity parameterisation that uses 4 parameters to describe a
    multivariate normal angular mosaic spread. Sigma is enforced as positive
    definite by parameterising using the cholesky decomposition.
    W = | w1  0  0  |
        | w2 w3  0  |
        | 0   0 w4 |
    S = W*W^T
    """

    @staticmethod
    def is_angular() -> bool:
        return True

    @staticmethod
    def num_parameters() -> int:
        return 4

    def sigma(self) -> np.array:
        """
        Compute the covariance matrix of the MVN from the parameters
        """
        # M = [[p[0], 0, 0], [p[1], p[2], 0], [0, 0, p[3]]]
        # return np.matmul(M, M.T)
        ab = self.params[0] * self.params[1]
        aa = self.params[0] ** 2
        bcsq = self.params[1] ** 2 + self.params[2] ** 2
        dd = self.params[3] ** 2
        return np.array([[aa, ab, 0.0], [ab, bcsq, 0], [0, 0, dd]], dtype=np.float64)

    def mosaicity(self) -> Dict:
        """Three components of mosaicity"""
        decomp = linalg.eigensystem.real_symmetric(
            matrix.sqr(flumpy.from_numpy(self.sigma())).as_flex_double_matrix()
        )
        m = mosaicity_from_eigen_decomposition(decomp.values())
        v = decomp.vectors()
        mosaicities = {"radial": 0, "angular_0": 0, "angular_1": 0}
        n_angular = 0
        for i in range(3):
            vec = (v[i * 3], v[(i * 3) + 1], v[(i * 3) + 2])
            if vec == (0, 0, 1):
                mosaicities["radial"] = m[i]
            else:
                mosaicities["angular_" + str(n_angular)] = m[i]
                n_angular += 1
        return mosaicities

    def first_derivatives(self) -> np.array:
        """
        Compute the first derivatives of Sigma w.r.t the parameters
        """
        b1, b2, b3, b4 = self.params

        # d1 = [[2 * b1, b2, 0], [b2, 0, 0], [0, 0, 0]]
        # d2 = [[0, b1, 0], [b1, 2 * b2, 0], [0, 0, 0]]
        # d3 = [[0, 0, 0], [0, 2 * b3, 0], [0, 0, 0]]
        # d4 = [[0, 0, 0], [0, 0, 0], [0, 0, 2 * b4]]
        ds = np.array(
            [
                [2 * b1, b2, 0, b2, 0, 0, 0, 0, 0],
                [0, b1, 0, b1, 2 * b2, 0, 0, 0, 0],
                [0, 0, 0, 0, 2 * b3, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 2 * b4],
            ],
            dtype=np.float64,
        ).reshape(4, 3, 3)
        return ds


class ModelState(object):
    """
    A class to keep track of the model state

    """

    def __init__(
        self,
        experiment,
        mosaicity_parameterisation,
        wavelength_parameterisation=None,
        fix_mosaic_spread=False,
        fix_wavelength_spread=True,
        fix_unit_cell=False,
        fix_orientation=False,
    ):
        """
        Initialise the model state

        """

        # Save the crystal model
        self.experiment = experiment
        self.crystal = experiment.crystal

        # The U and P parameterisation
        self._U_parameterisation = CrystalOrientationParameterisation(self.crystal)
        self._B_parameterisation = CrystalUnitCellParameterisation(self.crystal)

        # The M and L parameterisations
        self._M_parameterisation = mosaicity_parameterisation
        self._L_parameterisation = wavelength_parameterisation

        # Set the flags to fix parameters
        self._is_mosaic_spread_fixed = fix_mosaic_spread
        self._is_wavelength_spread_fixed = fix_wavelength_spread
        self._is_unit_cell_fixed = fix_unit_cell
        self._is_orientation_fixed = fix_orientation

        # Check wavelength parameterisation
        if not self.is_wavelength_spread_fixed:
            assert self._L_parameterisation is not None

    @property
    def is_orientation_fixed(self) -> bool:
        return self._is_orientation_fixed

    @property
    def is_unit_cell_fixed(self) -> bool:
        return self._is_unit_cell_fixed

    @property
    def is_mosaic_spread_fixed(self) -> bool:
        return self._is_mosaic_spread_fixed

    @property
    def is_mosaic_spread_angular(self) -> bool:
        return self._M_parameterisation.is_angular()

    @property
    def is_wavelength_spread_fixed(self) -> bool:
        return self._is_wavelength_spread_fixed

    @property
    def unit_cell(self):
        return self.crystal.get_unit_cell()

    @property
    def U_matrix(self) -> np.array:
        return np.array([self.crystal.get_U()], dtype=np.float64).reshape(3, 3)

    @property
    def B_matrix(self) -> np.array:
        return np.array([self.crystal.get_B()], dtype=np.float64).reshape(3, 3)

    @property
    def A_matrix(self) -> np.array:
        return np.array([self.crystal.get_A()], dtype=np.float64).reshape(3, 3)

    @property
    def mosaicity_covariance_matrix(self) -> np.array:
        return self._M_parameterisation.sigma()

    @property
    def wavelength_spread(self) -> flex.double:
        if self._L_parameterisation is not None:
            return self._L_parameterisation.sigma()
        return flex.double()

    @property
    def U_params(self) -> np.array:
        """Get the parameters of the orientation parameterisation"""
        return np.array(self._U_parameterisation.get_param_vals(), dtype=np.float64)

    @U_params.setter
    def U_params(self, params) -> None:
        self._U_parameterisation.set_param_vals(tuple(float(i) for i in params))

    @property
    def B_params(self) -> np.array:
        """Get the parameters of the orientation parameterisation"""
        return np.array(self._B_parameterisation.get_param_vals(), dtype=np.float64)

    @B_params.setter
    def B_params(self, params) -> None:
        self._B_parameterisation.set_param_vals(tuple(float(i) for i in params))

    @property
    def M_params(self) -> np.array:
        "Parameters of the mosaicity parameterisation"
        return self._M_parameterisation.parameters

    @M_params.setter
    def M_params(self, params) -> None:
        self._M_parameterisation.parameters = params

    @property
    def L_params(self) -> np.array:
        "Parameters of the Lambda (wavelength) parameterisation"
        if self._L_parameterisation is not None:
            return np.array(self._L_parameterisation.parameters, dtype=np.float64)
        return np.array([])

    @L_params.setter
    def L_params(self, params: flex.double) -> None:
        if self._L_parameterisation is not None:
            self._L_parameterisation.parameters = params

    @property
    def dU_dp(self) -> np.array:
        """
        Get the first derivatives of U w.r.t its parameters

        """
        ds_dp = self._U_parameterisation.get_ds_dp()
        n = len(ds_dp)
        s = np.array([ds_dp[i] for i in range(n)], dtype=np.float64).reshape(n, 3, 3)
        return s

    @property
    def dB_dp(self) -> np.array:
        """
        Get the first derivatives of B w.r.t its parameters

        """
        ds_dp = self._B_parameterisation.get_ds_dp()
        n = len(ds_dp)
        s = np.array([ds_dp[i] for i in range(n)], dtype=np.float64).reshape(n, 3, 3)
        return s

    @property
    def dM_dp(self) -> np.array:
        """
        Get the first derivatives of M w.r.t its parameters

        """
        return self._M_parameterisation.first_derivatives()

    @property
    def dL_dp(self) -> flex.double:
        """
        Get the first derivatives of L w.r.t its parameters

        """
        if self._L_parameterisation is not None:
            return self._L_parameterisation.first_derivatives()
        return flex.double()

    @property
    def active_parameters(self) -> np.array:
        """
        The active parameters in order: U, B, M, L, W
        """
        active_params = []
        if not self.is_orientation_fixed:
            active_params.append(self.U_params)
        if not self.is_unit_cell_fixed:
            active_params.append(self.B_params)
        if not self.is_mosaic_spread_fixed:
            active_params.append(self.M_params)
        if not self.is_wavelength_spread_fixed:
            active_params.append(self.L_params)
        active_params = np.concatenate(active_params)
        assert len(active_params) > 0
        return active_params

    @active_parameters.setter
    def active_parameters(self, params) -> None:
        """
        Set the active parameters in order: U, B, M, L, W
        """
        if not self.is_orientation_fixed:
            n_U_params = len(self.U_params)
            temp = params[:n_U_params]
            params = params[n_U_params:]
            self.U_params = temp
        if not self.is_unit_cell_fixed:
            n_B_params = len(self.B_params)
            temp = params[:n_B_params]
            params = params[n_B_params:]
            self.B_params = temp
        if not self.is_mosaic_spread_fixed:
            n_M_params = self.M_params.size
            temp = params[:n_M_params]
            params = params[n_M_params:]
            self.M_params = np.array(temp)
        if not self.is_wavelength_spread_fixed:
            n_L_params = self.L_params.size
            temp = params[:n_L_params]
            self.L_params = temp

    @property
    def parameter_labels(self) -> List[str]:
        """
        Get the parameter labels

        """
        labels = []
        if not self.is_orientation_fixed:
            labels += [f"Crystal_U_{i}" for i in range(self.U_params.size)]
        if not self.is_unit_cell_fixed:
            labels += [f"Crystal_B_{i}" for i in range(self.B_params.size)]
        if not self.is_mosaic_spread_fixed:
            labels += [f"Mosaicity_{i}" for i in range(self.M_params.size)]
        if not self.is_wavelength_spread_fixed:
            labels.append("Wavelength_Spread")
        assert len(labels) > 0
        return labels


class ReflectionModelState(object):
    """
    Class to compute basic derivatives of Sigma and r w.r.t parameters

    """

    def __init__(self, state, s0, h):
        """
        Initialise with the state and compute derivatives

        """

        # Compute the reciprocal lattice vector
        self._h = np.array(h, dtype=np.float64)
        self._r = np.einsum("ij,j->i", state.A_matrix, self._h)
        self._s0 = np.array(s0, dtype=np.float64)
        self._norm_s0 = (self._s0 / norm(self._s0)).flatten()

        self.state = state
        self._Q = None

        n_params = 0
        if not self.state.is_orientation_fixed:
            n_params += len(self.state.U_params)
        if not self.state.is_unit_cell_fixed:
            n_params += len(self.state.B_params)
        if not self.state.is_mosaic_spread_fixed:
            n_params += len(self.state.M_params)
        if not self.state.is_wavelength_spread_fixed:
            n_params += len(self.state.L_params)

        # The array of derivatives
        self._dr_dp = np.zeros(shape=(3, n_params), dtype=np.float64)
        self._ds_dp = np.zeros(shape=(3, 3, n_params), dtype=np.float64)
        self._dl_dp = flex.double(n_params)

        if self.state.is_mosaic_spread_angular:
            norm_r = self._r / norm(self._r)
            q1 = np.cross(norm_r, self._norm_s0)
            q1 /= norm(q1)
            q2 = np.cross(norm_r, q1)
            q2 /= norm(q2)
            self._Q = np.array([q1, q2, norm_r], dtype=np.float64).reshape(3, 3)
        self._recalc_sigma()
        self._recalc_sigma_lambda()
        self.update()

    def _recalc_sigma(self):
        # Compute the covariance matrix
        M = self.state.mosaicity_covariance_matrix
        if self.state.is_mosaic_spread_angular:
            # Define rotation for W sigma components
            # check if r has actually been updated
            if (not self.state.is_orientation_fixed) or (
                not self.state.is_unit_cell_fixed
            ):
                norm_r = self._r / norm(self._r)
                q1 = np.cross(norm_r, self._norm_s0)
                q1 /= norm(q1)
                q2 = np.cross(norm_r, q1)
                q2 /= norm(q2)
                self._Q = np.array([q1, q2, norm_r], dtype=np.float64).reshape(3, 3)
            self._sigma = np.matmul(np.matmul(self._Q.T, M), self._Q)
        else:
            self._sigma = M  #

    def _recalc_sigma_lambda(self):
        # Get the wavelength spread

        assert len(self.state.wavelength_spread) == len(self.state.dL_dp)
        if len(self.state.wavelength_spread) == 0:
            self._sigma_lambda = 0
        else:
            assert len(self.state.wavelength_spread) == 1
            self._sigma_lambda = self.state.wavelength_spread[0]

    def update(self):
        "Updates r, sigma, derivatives based on latest model state"

        # Set the reciprocal lattice vector
        if (not self.state.is_orientation_fixed) or (not self.state.is_unit_cell_fixed):
            self._r = np.matmul(self.state.A_matrix, self._h)
        if not self.state.is_wavelength_spread_fixed:
            self._recalc_sigma_lambda()
        if not self.state.is_mosaic_spread_fixed:
            self._recalc_sigma()

        # Compute derivatives w.r.t U parameters
        n_tot = 0
        state = self.state
        if not state.is_orientation_fixed:
            dU_dp = self.state.dU_dp
            n_U_params = dU_dp.shape[0]
            dUBh = np.einsum("lij,jk,k->il", dU_dp, state.B_matrix, self._h)
            self._dr_dp[:, n_tot : n_tot + n_U_params] = dUBh
            n_tot += n_U_params

        # Compute derivatives w.r.t B parameters
        if not state.is_unit_cell_fixed:
            dB_dp = self.state.dB_dp
            n_B_params = dB_dp.shape[0]  # state.B_params.size
            UdBh = np.einsum("ij,ljk,k->il", state.U_matrix, dB_dp, self._h)
            self._dr_dp[:, n_tot : n_tot + n_B_params] = UdBh
            n_tot += n_B_params

        # Compute derivatives w.r.t M parameters
        if not state.is_mosaic_spread_fixed:
            dM_dp = self.state.dM_dp
            n_M_params = dM_dp.shape[0]  # state.M_params.size
            if state.is_mosaic_spread_angular:
                QTMQ = np.einsum("ij,mjk,kl->ilm", self._Q.T, dM_dp, self._Q)
                self._ds_dp[:, :, n_tot : n_tot + n_M_params] = QTMQ
            else:
                self._ds_dp[:, :, n_tot : n_tot + n_M_params] = np.transpose(
                    dM_dp, axes=(1, 2, 0)
                )
            n_tot += n_M_params
        # Compute derivatives   w.r.t L parameters
        if not state.is_wavelength_spread_fixed:
            self._dl_dp[n_tot] = self.state.dL_dp[0]

    @property
    def mosaicity_covariance_matrix(self) -> np.array:
        return self._sigma

    def get_r(self) -> np.array:
        """
        Return the reciprocal lattice vector

        """
        return self._r.reshape(3, 1)

    def get_dS_dp(self) -> np.array:
        """
        Return the derivatives of the covariance matrix (an array of size 3x3xn)

        """
        return self._ds_dp

    def get_dr_dp(self) -> np.array:
        """
        Return the derivatives of the reciprocal lattice vector ( an array of size 3xn)

        """
        return self._dr_dp

    @property
    def wavelength_spread(self) -> float:
        return self._sigma_lambda

    def get_dL_dp(self) -> flex.double:
        """
        Return the derivatives of the wavelength spread

        """
        return self._dl_dp
